# -*- coding: utf-8 -*-
"""This module defines classes to represent the initial excitation."""

import numpy as np
import smuthi.fields as flds
import smuthi.fields.expansions as fldex
import smuthi.fields.transformations as trf
import smuthi.fields.vector_wave_functions as vwf
import smuthi.particles as part
import smuthi.linearsystem.particlecoupling.direct_coupling as dircoup
import smuthi.linearsystem.particlecoupling.layer_mediated_coupling as laycoup
import smuthi.postprocessing.scattered_field as sf
import smuthi.postprocessing.far_field as farf
import smuthi.utility.memoizing as memo
import warnings
import sys
from tqdm import tqdm


class InitialField:
    """Base class for initial field classes"""
    def __init__(self, vacuum_wavelength):
        self.vacuum_wavelength = vacuum_wavelength
        self.validity_conditions = []

    def spherical_wave_expansion(self, particle, layer_system):
        """Virtual method to be overwritten."""
        pass

    def plane_wave_expansion(self, layer_system, i):
        """Virtual method to be overwritten."""
        pass

    def piecewise_field_expansion(self, layer_system):
        empty_pfe = fldex.PiecewiseFieldExpansion()
        empty_pfe.validity_conditions += self.validity_conditions
        return empty_pfe

    def angular_frequency(self):
        """Angular frequency.

        Returns:
            Angular frequency (float) according to the vacuum wavelength in units of c=1.
        """
        return flds.angular_frequency(self.vacuum_wavelength)

    def get_k_parallel_array(self):
        """Get k_parallel array which is either the default array or the one stored in the object"""
        if not hasattr(self, 'k_parallel_array'):
            return flds.default_initial_field_k_parallel_array
        if type(self.k_parallel_array) == str and self.k_parallel_array == 'default':
            return flds.default_initial_field_k_parallel_array
        else:
            return self.k_parallel_array

    def get_azimuthal_angles_array(self):
        """Get azimuthal angles array which is either the default array or the one stored in the object"""
        if not hasattr(self, 'azimuthal_angles_array'):
            return flds.default_azimuthal_angles
        if type(self.azimuthal_angles_array) == str and self.azimuthal_angles_array == 'default':
            return flds.default_azimuthal_angles
        else:
            return self.azimuthal_angles_array


class InitialPropagatingWave(InitialField):
    """Base class for plane waves and Gaussian beams

    Args:
        vacuum_wavelength (float):
        polar_angle (float):            polar propagation angle (0 means, parallel to z-axis)
        azimuthal_angle (float):        azimuthal propagation angle (0 means, in x-z plane)
        polarization (int):             0 for TE/s, 1 for TM/p
        amplitude (float or complex):   Electric field amplitude
        reference_point (list):         Location where electric field of incoming wave equals amplitude
    """
    def __init__(self, vacuum_wavelength, polar_angle, azimuthal_angle, polarization, amplitude=1,
                 reference_point=None):
        assert (polarization == 0 or polarization == 1)
        InitialField.__init__(self, vacuum_wavelength)
        if np.isclose(np.cos(polar_angle), 0):
            raise ValueError('propagating waves not defined in the xy-plane')
        self.polar_angle = polar_angle
        self.azimuthal_angle = azimuthal_angle
        self.polarization = polarization
        self.amplitude = amplitude
        if reference_point:
            self.reference_point = reference_point
        else:
            self.reference_point = [0, 0, 0]

    def spherical_wave_expansion(self, particle, layer_system):
        """Regular spherical wave expansion of the wave including layer system response, at the locations of the
        particles.

        Args:
            particle (smuthi.particles.Particle):    particle relative to which the swe is computed
            layer_system (smuthi.layer.LayerSystem): stratified medium

        Returns:
            regular smuthi.field_expansion.SphericalWaveExpansion object
        """
        i = layer_system.layer_number(particle.position[2])
        pwe_up, pwe_down = self.plane_wave_expansion(layer_system, i)
        return (trf.pwe_to_swe_conversion(pwe_up, particle.l_max, particle.m_max, particle.position)
                + trf.pwe_to_swe_conversion(pwe_down, particle.l_max, particle.m_max, particle.position))

    def piecewise_field_expansion(self, layer_system):
        """Compute a piecewise field expansion of the initial field.

        Args:
            layer_system (smuthi.layer.LayerSystem):    stratified medium

        Returns:
            smuthi.field_expansion.PiecewiseWaveExpansion object
        """
        pfe = InitialField.piecewise_field_expansion(self, layer_system)
        for i in range(layer_system.number_of_layers()):
            pwe_up, pwe_down = self.plane_wave_expansion(layer_system, i)
            pfe.expansion_list.append(pwe_up)
            pfe.expansion_list.append(pwe_down)
        return pfe

    def electric_field(self, x, y, z, layer_system):
        """Evaluate the complex electric field corresponding to the wave.

        Args:
            x (array like):     Array of x-values where to evaluate the field (length unit)
            y (array like):     Array of y-values where to evaluate the field (length unit)
            z (array like):     Array of z-values where to evaluate the field (length unit)
            layer_system (smuthi.layer.LayerSystem):    Stratified medium

        Returns
            Tuple (E_x, E_y, E_z) of electric field values
        """

        pfe = self.piecewise_field_expansion(layer_system=layer_system)
        return pfe.electric_field(x, y, z)

    def magnetic_field(self, x, y, z, layer_system):
        """Evaluate the complex magnetic field corresponding to the wave.

        Args:
            x (array like):     Array of x-values where to evaluate the field (length unit)
            y (array like):     Array of y-values where to evaluate the field (length unit)
            z (array like):     Array of z-values where to evaluate the field (length unit)
            layer_system (smuthi.layer.LayerSystem):    Stratified medium

        Returns
            Tuple (H_x, H_y, H_z) of magnetic field values
        """

        pfe = self.piecewise_field_expansion(layer_system=layer_system)
        return pfe.magnetic_field(x, y, z, self.vacuum_wavelength)


class GaussianBeam(InitialPropagatingWave):
    """Class for the representation of a Gaussian beam as initial field."""
    def __init__(self, vacuum_wavelength, polar_angle, azimuthal_angle, polarization, beam_waist,
                 k_parallel_array='default', azimuthal_angles_array='default', amplitude=1, reference_point=None):
        InitialPropagatingWave.__init__(self, vacuum_wavelength, polar_angle, azimuthal_angle, polarization, amplitude,
                                        reference_point)
        self.beam_waist = beam_waist
        self.k_parallel_array = k_parallel_array
        self.azimuthal_angles_array = azimuthal_angles_array

    def plane_wave_expansion(self, layer_system, i, k_parallel_array=None, azimuthal_angles_array=None):
        """Plane wave expansion of the Gaussian beam.

        Args:
            layer_system (smuthi.layer.LayerSystem):    stratified medium
            i (int):                                    layer number in which to evaluate the expansion
            k_parallel_array (numpy.ndarray):           in-plane wavenumber array for the expansion. if none specified,
                                                        self.k_parallel_array is used
            azimuthal_angles_array (numpy.ndarray):     azimuthal angles for the expansion. if none specified,
                                                        self.azimuthal_angles_array is used

        Returns:
            tuple of to smuthi.field_expansion.PlaneWaveExpansion objects, one for upgoing and one for downgoing
            component
        """
        if k_parallel_array is None:
            k_parallel_array = InitialField.get_k_parallel_array(self)
        if azimuthal_angles_array is None:
            azimuthal_angles_array = InitialField.get_azimuthal_angles_array(self)

        if np.cos(self.polar_angle) > 0:
            iG = 0  # excitation layer number
            kind = 'upgoing'
        else:
            iG = layer_system.number_of_layers() - 1
            kind = 'downgoing'

        niG = layer_system.refractive_indices[iG]  # refractive index in excitation layer
        if niG.imag:
            warnings.warn('beam coming from absorbing medium')
        k_iG = niG * self.angular_frequency()
        z_iG = layer_system.reference_z(iG)
        loz = layer_system.lower_zlimit(iG)
        upz = layer_system.upper_zlimit(iG)
        pwe_exc = fldex.PlaneWaveExpansion(k=k_iG, k_parallel=k_parallel_array, azimuthal_angles=azimuthal_angles_array,
                                           kind=kind, reference_point=[0, 0, z_iG], lower_z=loz, upper_z=upz)

        k_Gx = k_iG * np.sin(self.polar_angle) * np.cos(self.azimuthal_angle)
        k_Gy = k_iG * np.sin(self.polar_angle) * np.sin(self.azimuthal_angle)

        kp = pwe_exc.k_parallel_grid()
        al = pwe_exc.azimuthal_angle_grid()

        kx = kp * np.cos(al)
        ky = kp * np.sin(al)
        kz = pwe_exc.k_z_grid()

        w = self.beam_waist
        r_G = self.reference_point

        g = (self.amplitude * w**2 / (4 * np.pi) * np.exp(-w**2 / 4 * ((kx - k_Gx)**2 + (ky - k_Gy)**2))
             * np.exp(-1j * (kx * r_G[0] + ky * r_G[1] + kz * (r_G[2] - z_iG))) )

        pwe_exc.coefficients[0, :, :] = g * np.cos(al - self.azimuthal_angle + self.polarization * np.pi/2)
        if np.cos(self.polar_angle) > 0:
            pwe_exc.coefficients[1, :, :] = g * np.sin(al - self.azimuthal_angle + self.polarization * np.pi/2)
        else:
            pwe_exc.coefficients[1, :, :] = - g * np.sin(al - self.azimuthal_angle + self.polarization * np.pi/2)

        pwe_up, pwe_down = layer_system.response(pwe_exc, from_layer=iG, to_layer=i)
        if iG == i:
            if kind == 'upgoing':
                pwe_up = pwe_up + pwe_exc
            elif kind == 'downgoing':
                pwe_down = pwe_down + pwe_exc

        return pwe_up, pwe_down

    def propagated_far_field(self, layer_system):
        """Evaluate the far field intensity of the reflected / transmitted initial field.

        Args:
            layer_system (smuthi.layers.LayerSystem):           Stratified medium

        Returns:
            A tuple of smuthi.field_expansion.FarField objects, one for forward (i.e., into the top hemisphere) and one
            for backward propagation (bottom hemisphere).
        """
        i_top = layer_system.number_of_layers() - 1
        top_ff = farf.pwe_to_ff_conversion(vacuum_wavelength=self.vacuum_wavelength,
                                            plane_wave_expansion=self.plane_wave_expansion(layer_system, i_top)[0])
        bottom_ff = farf.pwe_to_ff_conversion(vacuum_wavelength=self.vacuum_wavelength,
                                               plane_wave_expansion=self.plane_wave_expansion(layer_system, 0)[1])

        return top_ff, bottom_ff

    def initial_intensity(self, layer_system):
        """Evaluate the incoming intensity of the initial field.

        Args:
            layer_system (smuthi.layers.LayerSystem):           Stratified medium

        Returns:
            A smuthi.field_expansion.FarField object holding the initial intensity information.
        """
        if np.cos(self.polar_angle) > 0:  # bottom illumination
            ff = farf.pwe_to_ff_conversion(vacuum_wavelength=self.vacuum_wavelength,
                                            plane_wave_expansion=self.plane_wave_expansion(layer_system, 0)[0])
        else:  # top illumination
            i_top = layer_system.number_of_layers() - 1
            ff = farf.pwe_to_ff_conversion(vacuum_wavelength=self.vacuum_wavelength,
                                           plane_wave_expansion=self.plane_wave_expansion(layer_system, i_top)[1])
        return ff


class PlaneWave(InitialPropagatingWave):
    """Class for the representation of a plane wave as initial field.

    Args:
        vacuum_wavelength (float):
        polar_angle (float):            polar angle of k-vector (0 means, k is parallel to z-axis)
        azimuthal_angle (float):        azimuthal angle of k-vector (0 means, k is in x-z plane)
        polarization (int):             0 for TE/s, 1 for TM/p
        amplitude (float or complex):   Plane wave amplitude at reference point
        reference_point (list):         Location where electric field of incoming wave equals amplitude
    """

    def plane_wave_expansion(self, layer_system, i):
        """Plane wave expansion for the plane wave including its layer system response. As it already is a plane wave,
        the plane wave expansion is somehow trivial (containing only one partial wave, i.e., a discrete plane wave
        expansion).

        Args:
            layer_system (smuthi.layers.LayerSystem): Layer system object
            i (int): layer number in which the plane wave expansion is valid

        Returns:
            Tuple of smuthi.field_expansion.PlaneWaveExpansion objects. The first element is an upgoing PWE, whereas the
            second element is a downgoing PWE.
        """
        if np.cos(self.polar_angle) > 0:
            iP = 0
            kind = 'upgoing'
        else:
            iP = layer_system.number_of_layers() - 1
            kind = 'downgoing'

        niP = layer_system.refractive_indices[iP]
        neff = np.sin([self.polar_angle]) * niP
        alpha = np.array([self.azimuthal_angle])

        angular_frequency = flds.angular_frequency(self.vacuum_wavelength)
        k_iP = niP * angular_frequency
        k_Px = k_iP * np.sin(self.polar_angle) * np.cos(self.azimuthal_angle)
        k_Py = k_iP * np.sin(self.polar_angle) * np.sin(self.azimuthal_angle)
        k_Pz = k_iP * np.cos(self.polar_angle)
        z_iP = layer_system.reference_z(iP)
        amplitude_iP = self.amplitude * np.exp(-1j * (k_Px * self.reference_point[0] + k_Py * self.reference_point[1]
                                                      + k_Pz * (self.reference_point[2] - z_iP)))
        loz = layer_system.lower_zlimit(iP)
        upz = layer_system.upper_zlimit(iP)
        pwe_exc = fldex.PlaneWaveExpansion(k=k_iP, k_parallel=neff*angular_frequency, azimuthal_angles=alpha, kind=kind,
                                           reference_point=[0, 0, z_iP], lower_z=loz, upper_z=upz)
        pwe_exc.coefficients[self.polarization, 0, 0] = amplitude_iP
        pwe_up, pwe_down = layer_system.response(pwe_exc, from_layer=iP, to_layer=i)
        if iP == i:
            if kind == 'upgoing':
                pwe_up = pwe_up + pwe_exc
            elif kind == 'downgoing':
                pwe_down = pwe_down + pwe_exc

        return pwe_up, pwe_down


class DipoleSource(InitialField):
    """Class for the representation of a single point dipole source.

    Args:
        vacuum_wavelength (float):      vacuum wavelength (length units)
        dipole_moment (list or tuple):  (x, y, z)-coordinates of dipole moment vector
        position (list or tuple):       (x, y, z)-coordinates of dipole position
        k_parallel_array (numpy.ndarray or str):  In-plane wavenumber.
                                                  If 'default', use smuthi.fields.default_initial_field_k_parallel_array
        azimuthal_angles_array (numpy.ndarray or str):    Azimuthal angles for plane wave expansions
                                                          If 'default', use smuthi.fields.default_azimuthal_angles
    """
    def __init__(self, vacuum_wavelength, dipole_moment, position, k_parallel_array='default',
                 azimuthal_angles_array='default'):
        InitialField.__init__(self, vacuum_wavelength)
        self.dipole_moment = dipole_moment
        self.position = position
        self.k_parallel_array = k_parallel_array
        self.azimuthal_angles_array = azimuthal_angles_array

    def current(self):
        r"""The current density takes the form

        .. math::
            \mathbf{j}(\mathbf{r}) = \delta(\mathbf{r} - \mathbf{r}_D) \mathbf{j}_D,

        where :math:`\mathbf{j}_D = -j \omega \mathbf{\mu}`, :math:`\mathbf{r}_D` is the location of the dipole, :math:`\omega`
        is the angular frequency and :math:`\mathbf{\mu}` is the dipole moment.
        For further details, see 'Principles of nano optics' by Novotny and Hecht.

        Returns:
            List of [x, y, z]-components of current density vector :math:`\mathbf{j}_D`
        """
        return [- 1j * self.angular_frequency() * self.dipole_moment[i] for i in range(3)]

    def outgoing_spherical_wave_expansion(self, layer_system):
        """The dipole field as an expansion in spherical vector wave functions.

        Args:
            layer_system (smuthi.layers.LayerSystem):   stratified medium

        Returns:
            outgoing smuthi.field_expansion.SphericalWaveExpansion object
        """
        laynum = layer_system.layer_number(self.position[2])
        k = layer_system.refractive_indices[laynum] * self.angular_frequency()
        swe_out = fldex.SphericalWaveExpansion(k=k, l_max=1, m_max=1, kind='outgoing', reference_point=self.position,
                                               lower_z=layer_system.lower_zlimit(laynum),
                                               upper_z=layer_system.upper_zlimit(laynum))
        l = 1
        for tau in range(2):
            for m in range(-1, 2):
                ex, ey, ez = vwf.spherical_vector_wave_function(0, 0, 0, k, 1, tau, l, -m)
                b = 1j * k / np.pi * 1j * self.angular_frequency() * (ex * self.current()[0] + ey * self.current()[1]
                                                                      + ez * self.current()[2])
                swe_out.coefficients[flds.multi_to_single_index(tau, l, m, 1, 1)] = b

        return swe_out

    def spherical_wave_expansion(self, particle, layer_system):
        """Regular spherical wave expansion of the wave including layer system response, at the locations of the
        particles.

        Args:
            particle (smuthi.particles.Particle):    particle relative to which the swe is computed
            layer_system (smuthi.layer.LayerSystem): stratified medium

        Returns:
            regular smuthi.field_expansion.SphericalWaveExpansion object
        """
        virtual_particle = part.Particle(position=self.position, l_max=1, m_max=1)
        wd = dircoup.direct_coupling_block(vacuum_wavelength=self.vacuum_wavelength, receiving_particle=particle,
                                           emitting_particle=virtual_particle, layer_system=layer_system)
        wr = laycoup.layer_mediated_coupling_block(vacuum_wavelength=self.vacuum_wavelength, receiving_particle=particle,
                                                   emitting_particle=virtual_particle, layer_system=layer_system,
                                                   k_parallel=self.get_k_parallel_array())
        k = self.angular_frequency() * layer_system.refractive_indices[layer_system.layer_number(particle.position[2])]
        swe = fldex.SphericalWaveExpansion(k=k, l_max=particle.l_max, m_max=particle.m_max, kind='regular',
                                           reference_point=particle.position)
        swe.coefficients = np.dot(wd + wr, self.outgoing_spherical_wave_expansion(layer_system).coefficients)
        return swe

    def piecewise_field_expansion(self, layer_system, include_direct_field=True, include_layer_response=True):
        """Compute a piecewise field expansion of the dipole field.

        Args:
            layer_system (smuthi.layer.LayerSystem):    stratified medium
            include_direct_field (bool):                if True (default), the direct dipole field is included.
                                                        otherwise, only the layer response of the dipole field is
                                                        returned.
            include_layer_response (bool):              if True (default), the layer response of the dipole field is
                                                        included. otherwise, only the direct dipole field is
                                                        returned.


        Returns:
            smuthi.field_expansion.PiecewiseWaveExpansion object
        """
        pfe = InitialField.piecewise_field_expansion(self, layer_system)
        if include_direct_field:
            pfe.expansion_list.append(self.outgoing_spherical_wave_expansion(layer_system))
        if include_layer_response:
            for i in range(layer_system.number_of_layers()):
                # layer response as plane wave expansions
                pwe_up, pwe_down = trf.swe_to_pwe_conversion(swe=self.outgoing_spherical_wave_expansion(layer_system),
                                                            k_parallel=self.get_k_parallel_array(),
                                                            azimuthal_angles=self.get_azimuthal_angles_array(),
                                                            layer_system=layer_system, layer_number=i,
                                                            layer_system_mediated=True)
                if i > 0:
                    pfe.expansion_list.append(pwe_up)
                if i < layer_system.number_of_layers() - 1:
                    pfe.expansion_list.append(pwe_down)

        return pfe

    def electric_field(self, x, y, z, layer_system, include_direct_field=True, include_layer_response=True):
        """Evaluate the complex electric field of the dipole source.

        Args:
            x (array like):     Array of x-values where to evaluate the field (length unit)
            y (array like):     Array of y-values where to evaluate the field (length unit)
            z (array like):     Array of z-values where to evaluate the field (length unit)
            layer_system (smuthi.layer.LayerSystem):    Stratified medium
            include_direct_field (bool):                if True (default), the direct dipole field is included.
                                                        otherwise, only the layer response of the dipole field is
                                                        returned.
            include_layer_response (bool):              if True (default), the layer response of the dipole field is
                                                        included. otherwise, only the direct dipole field is
                                                        returned.


        Returns
            Tuple (E_x, E_y, E_z) of electric field values
        """
        pfe = self.piecewise_field_expansion(layer_system=layer_system, include_direct_field=include_direct_field,
                                             include_layer_response=include_layer_response)
        return pfe.electric_field(x, y, z)

    def magnetic_field(self, x, y, z, layer_system, include_direct_field=True, include_layer_response=True):
        """Evaluate the complex magnetic field of the dipole source.

        Args:
            x (array like):     Array of x-values where to evaluate the field (length unit)
            y (array like):     Array of y-values where to evaluate the field (length unit)
            z (array like):     Array of z-values where to evaluate the field (length unit)
            layer_system (smuthi.layer.LayerSystem):    Stratified medium
            include_direct_field (bool):                if True (default), the direct dipole field is included.
                                                        otherwise, only the layer response of the dipole field is
                                                        returned.
            include_layer_response (bool):              if True (default), the layer response of the dipole field is
                                                        included. otherwise, only the direct dipole field is
                                                        returned.


        Returns
            Tuple (H_x, H_y, H_z) of electric field values
        """
        pfe = self.piecewise_field_expansion(layer_system=layer_system, include_direct_field=include_direct_field,
                                             include_layer_response=include_layer_response)
        return pfe.magnetic_field(x, y, z, self.vacuum_wavelength)

    def dissipated_power_homogeneous_background(self, layer_system):
        r"""Compute the power that the dipole would radiate in an infinite homogeneous medium of the same refractive
        index as the layer that contains the dipole.

        .. math::
            P_0 = \frac{|\mathbf{\mu}| k \omega^3}{12 \pi}

        Args:
            layer_system (smuthi.layers.LayerSystem): stratified medium

        Returns:
            power (float)
        """
        laynum = layer_system.layer_number(self.position[2])
        k = layer_system.refractive_indices[laynum] * self.angular_frequency()
        mu2 = abs(self.dipole_moment[0])**2 + abs(self.dipole_moment[1])**2 + abs(self.dipole_moment[2])**2
        p = mu2 * k * self.angular_frequency()**3 / (12 * np.pi)
        return p

    def check_dissipated_power_homogeneous_background(self, layer_system):
        laynum = layer_system.layer_number(self.position[2])
        e_x_in, e_y_in, e_z_in = self.electric_field(x=self.position[0]+10, y=self.position[1]+10, z=self.position[2]+10,
                                                     layer_system=layer_system, include_direct_field=True)
        k = layer_system.refractive_indices[laynum] * self.angular_frequency()
        p = self.angular_frequency() / 2 * (np.conjugate(self.dipole_moment[0]) * (e_x_in)
                                            + np.conjugate(self.dipole_moment[1]) * (e_y_in)
                                            + np.conjugate(self.dipole_moment[2]) * (e_z_in)).imag
        return p

    def dissipated_power(self, particle_list, layer_system, show_progress=True):
        r"""Compute the power that the dipole feeds into the system.

        It is computed according to

        .. math::
            P = P_0 + \frac{\omega}{2} \mathrm{Im} (\mathbf{\mu}^* \cdot \mathbf{E}(\mathbf{r}_D))

        where :math:`P_0` is the power that the dipole would feed into an infinte homogeneous medium with the same
        refractive index as the layer that contains the dipole, :math:`\mathbf{r}_D` is the location of the dipole,
        :math:`\omega` is the angular frequency, :math:`\mathbf{\mu}` is the dipole moment and :math:`\mathbf{E}`
        includes the reflections of the dipole field from the layer interfaces, as well as the scattered field from all
        particles.

        Args:
            particle_list (list of smuthi.particles.Particle objects): scattering particles
            layer_system (smuthi.layers.LayerSystem):                  stratified medium
            show_progress (bool):                                      if true, display progress

        Returns:
            dissipated power as float
        """

        k = layer_system.wavenumber(layer_system.layer_number(self.position[2]), self.vacuum_wavelength)
        virtual_particle = part.Particle(position=self.position, l_max=1, m_max=1)
        scattered_field_swe = fldex.SphericalWaveExpansion(k=k, l_max=1, m_max=1, kind='regular',
                                                           reference_point=self.position)

        if show_progress:
            iterator = tqdm(particle_list, desc='Dipole dissipated power   ', file=sys.stdout,
                            bar_format='{l_bar}{bar}| elapsed: {elapsed} remaining: {remaining}')
        else:
            iterator = particle_list

        for particle in iterator:
            wd = dircoup.direct_coupling_block(vacuum_wavelength=self.vacuum_wavelength,
                                        receiving_particle=virtual_particle,
                                        emitting_particle=particle,
                                        layer_system=layer_system)

            wr = laycoup.layer_mediated_coupling_block(vacuum_wavelength=self.vacuum_wavelength,
                                                receiving_particle=virtual_particle,
                                                emitting_particle=particle,
                                                layer_system=layer_system,
                                                k_parallel=self.get_k_parallel_array())

            scattered_field_swe.coefficients += np.dot(wd + wr, particle.scattered_field.coefficients)

        e_x_scat, e_y_scat, e_z_scat = scattered_field_swe.electric_field(x=self.position[0],
                                                                          y=self.position[1],
                                                                          z=self.position[2])

        e_x_in, e_y_in, e_z_in = self.electric_field(x=self.position[0], y=self.position[1], z=self.position[2],
                                                     layer_system=layer_system, include_direct_field=False)

        power = self.angular_frequency() / 2 * (np.conjugate(self.dipole_moment[0]) * (e_x_scat + e_x_in)
                                                + np.conjugate(self.dipole_moment[1]) * (e_y_scat + e_y_in)
                                                + np.conjugate(self.dipole_moment[2]) * (e_z_scat + e_z_in)).imag
        return self.dissipated_power_homogeneous_background(layer_system) + power

    def dissipated_power_alternative(self, particle_list, layer_system):
        r"""Compute the power that the dipole feeds into the system.

        It is computed according to

        .. math::
            P = P_0 + \frac{\omega}{2} \mathrm{Im} (\mathbf{\mu}^* \cdot \mathbf{E}(\mathbf{r}_D))

        where :math:`P_0` is the power that the dipole would feed into an infinte homogeneous medium with the same
        refractive index as the layer that contains the dipole, :math:`\mathbf{r}_D` is the location of the dipole,
        :math:`\omega` is the angular frequency, :math:`\mathbf{\mu}` is the dipole moment and :math:`\mathbf{E}`
        includes the reflections of the dipole field from the layer interfaces, as well as the scattered field from all
        particles. In contrast to dissipated_power, this routine relies on the scattered field piecewise expansion and
        and might thus be slower.

        Args:
            particle_list (list of smuthi.particles.Particle objects): scattering particles
            layer_system (smuthi.layers.LayerSystem): stratified medium

        Returns:
            dissipated power as float
        """
        scat_fld_exp = sf.scattered_field_piecewise_expansion(self.vacuum_wavelength, particle_list, layer_system,
                                                              self.get_k_parallel_array(),
                                                              self.get_azimuthal_angles_array())
        e_x_scat, e_y_scat, e_z_scat = scat_fld_exp.electric_field(self.position[0], self.position[1], self.position[2])
        e_x_in, e_y_in, e_z_in = self.electric_field(x=self.position[0], y=self.position[1], z=self.position[2],
                                                     layer_system=layer_system, include_direct_field=False)
        power = self.angular_frequency() / 2 * (np.conjugate(self.dipole_moment[0]) * (e_x_scat + e_x_in)
                                                + np.conjugate(self.dipole_moment[1]) * (e_y_scat + e_y_in)
                                                + np.conjugate(self.dipole_moment[2]) * (e_z_scat + e_z_in)).imag
        return self.dissipated_power_homogeneous_background(layer_system) + power

    def plane_wave_expansion(self, layer_system, i, k_parallel_array=None, azimuthal_angles_array=None):
        """Plane wave expansion of the dipole field.

        Args:
            layer_system (smuthi.layer.LayerSystem):    stratified medium
            i (int):                                    layer number in which to evaluate the expansion
            k_parallel_array (numpy.ndarray):           in-plane wavenumber array for the expansion. if none specified,
                                                        self.k_parallel_array is used
            azimuthal_angles_array (numpy.ndarray):     azimuthal angles for the expansion. if none specified,
                                                        self.azimuthal_angles_array is used

        Returns:
            tuple of to smuthi.field_expansion.PlaneWaveExpansion objects, one for upgoing and one for downgoing
            component
        """
        if k_parallel_array is None:
            k_parallel_array = self.get_k_parallel_array()
        if azimuthal_angles_array is None:
            azimuthal_angles_array = self.get_azimuthal_angles_array()

        virtual_particle = part.Particle(position=self.position)
        virtual_particle.scattered_field = self.outgoing_spherical_wave_expansion(layer_system)

        return sf.scattered_field_pwe(self.vacuum_wavelength, [virtual_particle], layer_system, i, k_parallel_array,
                                      azimuthal_angles_array, include_direct=True, include_layer_response=True)


class DipoleCollection(InitialField):
    """Class for the representation of a set of point dipole sources. Use the append method to add DipoleSource objects.

    Args:
        vacuum_wavelength (float):                      vacuum wavelength (length units)
        k_parallel_array (numpy.ndarray or str):        In-plane wavenumber.
                                                        If 'default', use smuthi.fields.default_initial_field_k_parallel_array
        azimuthal_angles_array (numpy.ndarray or str):  Azimuthal angles for plane wave expansions
                                                        If 'default', use smuthi.fields.default_azimuthal_angles
        angular_resolution (float):                     If provided, angular arrays are generated with this angular resolution
                                                        over the default angular range
        compute_swe_by_pwe (bool):    If True, the initial field coefficients are computed through a plane wave
                                      expansion of the whole dipole collection field. This is slower for few dipoles
                                      and particles, but can become faster than the default for many dipoles and
                                      particles (default=False).
        compute_dissipated_power_by_pwe (bool): If True, evaluate dissipated power through a plane wave expansion of the
                                                whole scattered field. This is slower for few dipoles,  but can be
                                                faster than the default for many dipoles (default=False).
    """
    def __init__(self, vacuum_wavelength, k_parallel_array='default', azimuthal_angles_array='default', angular_resolution=None,
                 compute_swe_by_pwe=False, compute_dissipated_power_by_pwe=False):
        InitialField.__init__(self, vacuum_wavelength=vacuum_wavelength)
        self.dipole_list = []
        self.compute_swe_by_pwe = compute_swe_by_pwe
        self.compute_dissipated_power_by_pwe = compute_dissipated_power_by_pwe

        self.k_parallel_array = k_parallel_array

        if angular_resolution is not None:
            self.azimuthal_angles_array, _ = flds.angular_arrays(angular_resolution)
        if type(azimuthal_angles_array) == str and azimuthal_angles_array == 'default':
            self.azimuthal_angles_array = flds.default_azimuthal_angles

    def append(self, dipole):
        """Add dipole to collection.

        Args:
            dipole (DipoleSource):    Dipole object to add.
        """
        assert dipole.vacuum_wavelength == self.vacuum_wavelength
        self.dipole_list.append(dipole)

    def spherical_wave_expansion(self, particle, layer_system):
        """Regular spherical wave expansion of the dipole collection including layer system response, at the locations
        of the particles. If self.compute_swe_by_pwe is True, use the dipole collection plane wave expansion, otherwise
        use the individual dipoles spherical_wave_expansion method.

        Args:
            particle (smuthi.particles.Particle):    particle relative to which the swe is computed
            layer_system (smuthi.layer.LayerSystem): stratified medium

        Returns:
            regular smuthi.field_expansion.SphericalWaveExpansion object
        """
        # compute by pwe?
        if self.compute_swe_by_pwe and not any(
            layer_system.layer_number(particle.position[2])
                == np.array([layer_system.layer_number(dip.position[2]) for dip in self.dipole_list])):
            # (make sure the particle is not in the same layer as one of the dipoles)
            return InitialPropagatingWave.spherical_wave_expansion(self, particle, layer_system)

        # otherwise, compute by swe of virtual particles
        k = self.angular_frequency() * layer_system.refractive_indices[layer_system.layer_number(particle.position[2])]
        swe = fldex.SphericalWaveExpansion(k=k, l_max=particle.l_max, m_max=particle.m_max, kind='regular',
                                           reference_point=particle.position)
        for dipole in self.dipole_list:
            swe = swe + dipole.spherical_wave_expansion(particle, layer_system)
        return swe

    def piecewise_field_expansion(self, layer_system):
        """Compute a piecewise field expansion of the dipole collection..

        Args:
            layer_system (smuthi.layer.LayerSystem):    stratified medium

        Returns:
            smuthi.field_expansion.PiecewiseWaveExpansion object
        """
        pfe = InitialField.piecewise_field_expansion(self, layer_system)
        for dipole in self.dipole_list:
            pfe = pfe + dipole.piecewise_field_expansion(layer_system, include_direct_field=True)

        return pfe

    def electric_field(self, x, y, z, layer_system):
        """Evaluate the complex electric field of the dipole collection.

        Args:
            x (array like):     Array of x-values where to evaluate the field (length unit)
            y (array like):     Array of y-values where to evaluate the field (length unit)
            z (array like):     Array of z-values where to evaluate the field (length unit)
            layer_system (smuthi.layer.LayerSystem):    Stratified medium

        Returns
            Tuple (E_x, E_y, E_z) of electric field values
        """
        pfe = self.piecewise_field_expansion(layer_system=layer_system)
        return pfe.electric_field(x, y, z)

    def magnetic_field(self, x, y, z, layer_system):
        """Evaluate the complex magnetic field of the dipole collection.

        Args:
            x (array like):     Array of x-values where to evaluate the field (length unit)
            y (array like):     Array of y-values where to evaluate the field (length unit)
            z (array like):     Array of z-values where to evaluate the field (length unit)
            layer_system (smuthi.layer.LayerSystem):    Stratified medium

        Returns
            Tuple (H_x, H_y, H_z) of magnetic field values
        """
        pfe = self.piecewise_field_expansion(layer_system=layer_system)
        return pfe.magnetic_field(x, y, z, self.vacuum_wavelength)

    def dissipated_power(self, particle_list, layer_system, k_parallel='default',
                         azimuthal_angles='default', angular_resolution=None):
        r"""Compute the power that the dipole collection feeds into the system.

        It is computed according to

        .. math::
            P = \sum_i P_{0, i} + \frac{\omega}{2} \mathrm{Im} (\mathbf{\mu}_i^* \cdot \mathbf{E}_i(\mathbf{r}_i))

        where :math:`P_{0,i}` is the power that the i-th dipole would feed into an infinte homogeneous medium with the
        same refractive index as the layer that contains that dipole, :math:`\mathbf{r}_i` is the location of the i-th
        dipole, :math:`\omega` is the angular frequency, :math:`\mathbf{\mu}_i` is the dipole moment and
        :math:`\mathbf{E}_i` includes the reflections of the dipole field from the layer interfaces, as well as the
        scattered field from all particles and the fields from all other dipoles.
        In contrast to dissipated_power_alternative, this routine uses the particle coupling routines and might be
        faster for many particles and few dipoles.

        Args:
            particle_list (list of smuthi.particles.Particle objects): scattering particles
            layer_system (smuthi.layers.LayerSystem): stratified medium
            k_parallel (ndarray or str): array of in-plane wavenumbers for plane wave expansions. If 'default', use
                                         smuthi.fields.default_initial_field_k_parallel_array
            azimuthal_angles (ndarray or str):  array of azimuthal angles for plane wave expansions. If 'default', use
                                                smuthi.fields.default_azimuthal_angles
            angular_resolution (float):         If provided, angular arrays are generated with this angular resolution
                                                over the default angular range

        Returns:
            dissipated power of each dipole (list of floats)
        """

        if self.compute_dissipated_power_by_pwe:
            return self.dissipated_power_alternative(particle_list,
                                                     layer_system,
                                                     k_parallel,
                                                     azimuthal_angles,
                                                     angular_resolution)

        sys.stdout.write('Dipole array dissipated power evaluation:\n')
        sys.stdout.flush()

        power_list = []

        x_positions = np.array([dipole.position[0] for dipole in self.dipole_list])
        y_positions = np.array([dipole.position[1] for dipole in self.dipole_list])
        z_positions = np.array([dipole.position[2] for dipole in self.dipole_list])

        e_x_in_list, e_y_in_list, e_z_in_list = [], [], []

        for dipole in tqdm(self.dipole_list, desc='Cross contribution        ', file=sys.stdout,
                           bar_format='{l_bar}{bar}| elapsed: {elapsed} remaining: {remaining}'):
            with warnings.catch_warnings():
                warnings.filterwarnings("ignore", message="invalid value encountered in multiply")
                warnings.filterwarnings("ignore", message="divide by zero encountered in true_divide")
                warnings.filterwarnings("ignore", message="invalid value encountered in true_divide")

                e_x_in, e_y_in, e_z_in = dipole.electric_field(x=x_positions,
                                                               y=y_positions,
                                                               z=z_positions,
                                                               layer_system=layer_system)
            e_x_in_list.append(e_x_in)
            e_y_in_list.append(e_y_in)
            e_z_in_list.append(e_z_in)

        for i, dipole in enumerate(tqdm(self.dipole_list, desc='Self and particle contr.  ', file=sys.stdout,
                                   bar_format='{l_bar}{bar}| elapsed: {elapsed} remaining: {remaining}')):
            e_x_in = sum([e_x_in[i] for i_other, e_x_in in enumerate(e_x_in_list) if i_other != i])
            e_y_in = sum([e_y_in[i] for i_other, e_y_in in enumerate(e_y_in_list) if i_other != i])
            e_z_in = sum([e_z_in[i] for i_other, e_z_in in enumerate(e_z_in_list) if i_other != i])

            p = (dipole.dissipated_power(particle_list, layer_system) +
                 dipole.angular_frequency() / 2 * (np.conjugate(dipole.dipole_moment[0]) * e_x_in +
                                                   np.conjugate(dipole.dipole_moment[1]) * e_y_in +
                                                   np.conjugate(dipole.dipole_moment[2]) * e_z_in).imag)
            power_list.append(p)

        return power_list

    def dissipated_power_alternative(self, particle_list, layer_system, k_parallel='default',
                                     azimuthal_angles='default', angular_resolution=None):
        r"""Compute the power that the dipole collection feeds into the system.

        It is computed according to

        .. math::
            P = \sum_i P_{0, i} + \frac{\omega}{2} \mathrm{Im} (\mathbf{\mu}_i^* \cdot \mathbf{E}_i(\mathbf{r}_i))

        where :math:`P_{0,i}` is the power that the i-th dipole would feed into an infinte homogeneous medium with the
        same refractive index as the layer that contains that dipole, :math:`\mathbf{r}_i` is the location of the i-th
        dipole, :math:`\omega` is the angular frequency, :math:`\mathbf{\mu}_i` is the dipole moment and
        :math:`\mathbf{E}_i` includes the reflections of the dipole field from the layer interfaces, as well as the
        scattered field from all particles and the fields from all other dipoles. In contrast to dissipated_power, this
        routine uses the scattered field piecewise expansion and might be faster for few particles or many dipoles.

        Args:
            particle_list (list of smuthi.particles.Particle objects): scattering particles
            layer_system (smuthi.layers.LayerSystem): stratified medium
            k_parallel (ndarray or str): array of in-plane wavenumbers for plane wave expansions. If 'default', use
                                         smuthi.fields.default_initial_field_k_parallel_array
            azimuthal_angles (ndarray or str):  array of azimuthal angles for plane wave expansions. If 'default', use
                                                smuthi.fields.default_azimuthal_angles
            angular_resolution (float):         If provided, angular arrays are generated with this angular resolution
                                                over the default angular range

        Returns:
            dissipated power of each dipole (list of floats)
        """
        sys.stdout.write('Dipole array dissipated power evaluation:\n')
        sys.stdout.flush()

        if type(k_parallel) == str and k_parallel == 'default':
            k_parallel = flds.default_initial_field_k_parallel_array

        if angular_resolution is not None:
            azimuthal_angles, _ = flds.angular_arrays(angular_resolution)
        if type(azimuthal_angles) == str and azimuthal_angles == 'default':
            azimuthal_angles = flds.default_azimuthal_angles

        power_list = []

        x_positions = np.array([dipole.position[0] for dipole in self.dipole_list])
        y_positions = np.array([dipole.position[1] for dipole in self.dipole_list])
        z_positions = np.array([dipole.position[2] for dipole in self.dipole_list])

        layer_numbers = set([layer_system.layer_number(z) for z in z_positions])
        scat_fld_exp = sf.scattered_field_piecewise_expansion(self.vacuum_wavelength, particle_list, layer_system,
                                                              k_parallel, azimuthal_angles, angular_resolution, layer_numbers)

        e_x_scat, e_y_scat, e_z_scat = scat_fld_exp.electric_field(x_positions, y_positions, z_positions)

        e_x_in_direct_list, e_y_in_direct_list, e_z_in_direct_list = [], [], []
        e_x_in_response_list, e_y_in_response_list, e_z_in_response_list = [], [], []

        for dipole in tqdm(self.dipole_list, desc='Self and cross contrib.   ', file=sys.stdout,
                           bar_format='{l_bar}{bar}| elapsed: {elapsed} ' 'remaining: {remaining}'):
            with warnings.catch_warnings():
                warnings.filterwarnings("ignore", message="invalid value encountered in multiply")
                warnings.filterwarnings("ignore", message="divide by zero encountered in true_divide")
                warnings.filterwarnings("ignore", message="invalid value encountered in true_divide")
                e_x_in_direct, e_y_in_direct, e_z_in_direct = dipole.electric_field(
                    x=x_positions, y=y_positions, z=z_positions, layer_system=layer_system, include_layer_response=False)

            e_x_in_response, e_y_in_response, e_z_in_response = dipole.electric_field(
                x=x_positions, y=y_positions, z=z_positions, layer_system=layer_system, include_direct_field=False)

            e_x_in_direct_list.append(e_x_in_direct)
            e_y_in_direct_list.append(e_y_in_direct)
            e_z_in_direct_list.append(e_z_in_direct)

            e_x_in_response_list.append(e_x_in_response)
            e_y_in_response_list.append(e_y_in_response)
            e_z_in_response_list.append(e_z_in_response)

        for i, dipole in enumerate(self.dipole_list):
            e_x_in = (sum([e_x_in[i] for i_other, e_x_in in enumerate(e_x_in_direct_list) if i_other != i])
                      + sum([e_x_in[i] for i_other, e_x_in in enumerate(e_x_in_response_list)]))
            e_y_in = (sum([e_y_in[i] for i_other, e_y_in in enumerate(e_y_in_direct_list) if i_other != i])
                      + sum([e_y_in[i] for i_other, e_y_in in enumerate(e_y_in_response_list)]))
            e_z_in = (sum([e_z_in[i] for i_other, e_z_in in enumerate(e_z_in_direct_list) if i_other != i])
                      + sum([e_z_in[i] for i_other, e_z_in in enumerate(e_z_in_response_list)]))

            p = (dipole.dissipated_power_homogeneous_background(layer_system) +
                 dipole.angular_frequency() / 2 * (np.conjugate(dipole.dipole_moment[0]) * (e_x_scat[i] + e_x_in) +
                                                   np.conjugate(dipole.dipole_moment[1]) * (e_y_scat[i] + e_y_in) +
                                                   np.conjugate(dipole.dipole_moment[2]) * (e_z_scat[i] + e_z_in)).imag)
            power_list.append(p)

        return power_list

    @memo.Memoize
    def plane_wave_expansion(self, layer_system, i, k_parallel_array=None, azimuthal_angles_array=None):
        """Plane wave expansion of the dipole collection's field.

        Args:
            layer_system (smuthi.layer.LayerSystem):    stratified medium
            i (int):                                    layer number in which to evaluate the expansion
            k_parallel_array (numpy.ndarray):           in-plane wavenumber array for the expansion
            azimuthal_angles_array (numpy.ndarray):     azimuthal angles for the expansion

        Returns:
            tuple of to smuthi.field_expansion.PlaneWaveExpansion objects, one for upgoing and one for downgoing
            component
        """
        if k_parallel_array is None:
            k_parallel_array = self.get_k_parallel_array()
        if azimuthal_angles_array is None:
            azimuthal_angles_array = self.get_azimuthal_angles_array()

        virtual_particle_list = []

        for dipole in self.dipole_list:
            virtual_particle = part.Particle(position=dipole.position)
            virtual_particle.scattered_field = dipole.outgoing_spherical_wave_expansion(layer_system)
            virtual_particle_list.append(virtual_particle)

        return sf.scattered_field_pwe(self.vacuum_wavelength, virtual_particle_list, layer_system, i, k_parallel_array,
                                      azimuthal_angles_array, include_direct=True, include_layer_response=True)



"""Functions for the translation of spherical vector wave functions."""

import smuthi.fields as flds
import smuthi.utility.math as sma
import smuthi.utility.memoizing as memo
import smuthi.fields.transformations as trf
import numpy as np
import sys

##############################################################################
#                       Highest Level Translation Method 
#                   Used to directely interface with Smuthi. 
#          All other funcitons in this file are internal to this method.
############################################################################## 
# Note: The direct coupling block is just an outgoing to incoming translation


def svwf_translate(k, d, lmax1, mmax1, lmax2, mmax2, kind, threaded = False):
    
    # Cython has a non-intuitive convention to dealing with strings/chars between python2 and python3
    # To avoid this, convert the readable string flag to a binary int flag. This prevents compatibility problems.
    # Find more at https://cython.readthedocs.io/en/latest/src/tutorial/strings.html
    if kind == 'outgoing to regular':
        ckind = 0 
    elif kind == 'outgoing to outgoing':
        ckind = 1
    else:
        raise Exception('''
                        The Spherical harmonic translations only support the following conversions:
                                            "outgoing to regular"
                                                    and 
                                            "outgoing to outgoing"
                        You specified a kind that was not one of those two strings.
                        ''')
        
    # If the translation kind is valid, precalculate necessary inputs for translation.
    # Note: Translations go from 1 -> 2
    blocksize1 = int(flds.blocksize(lmax1, mmax1)) # Rows of translation matrix
    blocksize2 = int(flds.blocksize(lmax2, mmax2)) # Cols of translation matrix
    w = np.zeros((blocksize1, blocksize2), dtype = np.complex128, order = 'C')
    sph = np.zeros((lmax1+lmax2+1), dtype=np.complex128, order = 'C')
    dx = float(d[0])
    dy = float(d[1])
    dz = float(d[2])
    
    #Note: It is allways better to use the hash tables so just do it. 
    if dz == 0:
        # Note: The hash table is memoized. So if this lmax and mmax combo was used 
        # prior (such as in the solution of the linear system) it will not be re-calculated
        a5leg_array, b5leg_array =  ab5_coefficient_and_legendre_hash_table(lmax1, lmax2, mmax1, mmax2)
        return svwf_translate_2D_from_hash_table(blocksize1, blocksize2,
                                                    w, sph, k, dx, dy,dz,
                                                    lmax1, lmax2, mmax1, mmax2,
                                                    a5leg_array, b5leg_array,                              
                                                    ckind, threaded)
    else:
        # Note: The hash table is memoized. So if this lmax and mmax combo was used 
        # prior (such as in the solution of the linear system) it will not be re-calculated
        a5_array, b5_array =  ab5_coefficient_hash_table(lmax1, lmax2, mmax1, mmax2)
        return svwf_translate_3D_from_hash_table(blocksize1, blocksize2,
                                                    w, sph, k, dx, dy,dz,
                                                    lmax1, lmax2, mmax1, mmax2,
                                                    a5_array, b5_array,                              
                                                    ckind, threaded)           
    



##############################################################################
#  Make  hash table of ab5  for a particle order pair 
#  This is for 3D coupling (Cython uses the GIL)
##############################################################################
# """
# This hash table is particularly good when all particles have the same order. 
# """

try:
    from smuthi.utility.cython.cython_speedups  import _ab5_coefficient_hash_table
    @memo.Memoize
    def  ab5_coefficient_hash_table(lmax1, lmax2, mmax1, mmax2):
        return _ab5_coefficient_hash_table(int(lmax1), int(lmax2), int(mmax1), int(mmax2))  
except:
    @memo.Memoize
    def  ab5_coefficient_hash_table(lmax1, lmax2, mmax1, mmax2):
        r"""Creates a hash table of the elements
        :math:`a5(l_1,m_1,l_2,m_2,l_d)` and :math:`a5(l_1,m_1,l_2,m_2,l_d)`
        found in appendix B of [Egel 2018 diss],where a5 and b5 are the coefficients used in the evaluation of the SVWF translation
        operator. This hash table is usefull in reducing computation time when calculating the coupling between two particles
        that exist in the same layer but do not have the same z coordinate.


        Args:
            lmax1 (int):           Largest polar quantum number of the recieving particle
            lmax2  (int):          Largest polar quantum number  of the emitting particle
            mmax1  (int):          Largest azimuthal quantum number of the recieving particle
            mmax2  (int):          Largest azimuthal quantum number of the emitting particle

    
        Returns:
            a5_array (ndarray):          Hash table of the elements in :math:`A` not dependent on :math:`/theta`, :math:`/phi` or :math:`kd`
            b5_array (ndarray):          Hash table of the elements in :math:`B` not dependent on :math:`/theta`, :math:`/phi` or :math:`kd`          

        """        
        a5_array = []
        b5_array = []
        #ab5_key = []
        for im1, m1 in enumerate(range(-mmax1, mmax1 + 1)):
            for im2, m2 in enumerate(range(-mmax2, mmax2 + 1)):
                for il1, l1 in enumerate(range(max(1, abs(m1)), lmax1 + 1)):
                    for il2, l2 in enumerate(range(max(1, abs(m2)), lmax2 + 1)):
                        for ild, ld in enumerate(range(max(abs(l1 - l2), abs(m1 - m2)), l1 + l2 + 1)): 
                            a5, b5 = trf.ab5_coefficients(l2, m2, l1, m1, ld)
                            #ab5_key.append([l1, l2, m1, m2, ld])
                            a5_array.append(a5)
                            b5_array.append(b5)
                          
        return a5_array, b5_array

##############################################################################
#         3D translation and direct coupling block using hash table 
#                          (Cython releases the Gil)
############################################################################## 


# Call to lower level translation funciton. Use Try to enable Cython if able. 
try:
    from smuthi.utility.cython.cython_speedups import svwf_translate_3D_from_hash_table
except:  
    def svwf_translate_3D_from_hash_table(blocksize1,blocksize2,
                                        w,sph,k,dx, dy, dz,
                                        lmax1, lmax2, mmax1, mmax2,
                                        a5_hash_table, b5_hash_table,
                                        kind, threaded):
        
        r"""Subroutine to calculate the translation between two locations
            that are in the same layer and do not have the same z coordinate. This 
            subroutine is called internally when no Cython accelorations are found.
            This subroutine is the Python equivalent of the Cython accelorated code.


        Args:
            blocksize1 (int):           Number of columns in the direct coupling block
            blocksize2 (int):           Number of rows in the direct coupling block
            w  (ndarray):               Zero initialized direct coupling matrix block as numpy array
            sph  (ndarray):             Hankel or Bessel function matrix as numpy array
            k (complex):                Wavenumber in the shared media
            dx (float):                 x-coordinate of the distance between the two particles
            dy  (float):                y-coordinate of the distance between the two particles
            dz  (float):                z-coordinate of the distance between the two particles
            lmax1 (int):           Largest polar quantum number of the recieving particle
            lmax2  (int):          Largest polar quantum number  of the emitting particle
            mmax1  (int):          Largest azimuthal quantum number of the recieving particle
            mmax2  (int):          Largest azimuthal quantum number of the emitting particle
            a5_array (ndarray):          Hash table of the elements in :math:`A` not dependent on :math:`/theta`, :math:`/phi` or :math:`kd`
            b5_array (ndarray):          Hash table of the elements in :math:`B` not dependent on :math:`/theta`, :math:`/phi` or :math:`kd`          
            threaded (bool):          Flag to enable multithreading (valid only for Cython where the Gil can be released). Currently hard coded to False.  
            
        Returns:
            w  (ndarray):          Direct coupling matrix block as numpy array.
        """        

        d = np.sqrt(dx**2 + dy**2 + dz**2)
        cos_theta = dz / d
        sin_theta = np.sqrt(dx**2 + dy**2) / d
        phi = np.arctan2(dy, dx)
    
        legendre = sma.legendre_normalized(cos_theta, sin_theta, lmax1 + lmax2)[0][:,:,0]
        
        if kind == 0: # 0 = outgoing to incoming wave, 1 = outgoing to outgoing wave. 
        #NOTE: an long flag is used instead of a string because Cython 
        # has troublesome conventions about datatyping strings from python v2 to v3
        # See more on this at https://cython.readthedocs.io/en/latest/src/tutorial/strings.html
            for n in range(lmax1 + lmax2 + 1):
                sph[n] = sma.spherical_hankel(n, k*d) 
        else:
            for n in range(lmax1 + lmax2 + 1):
                sph[n] = sma.spherical_bessel(n, k*d) 
                
        # the particle coupling operator is the transpose of the SVWF translation operator
        # therefore, (l1,m1) and (l2,m2) are interchanged:
        idx_ab5 = 0
        for m1 in range(-mmax1, mmax1 + 1):
            for m2 in range(-mmax2, mmax2 + 1):
                eimph_val = np.exp(1j * (m2 - m1) * phi)
                for l1 in range(max(1, abs(m1)), lmax1 + 1):
                    for l2 in range(max(1, abs(m2)), lmax2 + 1):
                        A, B = complex(0), complex(0)
                        for ld in range(max(abs(l1 - l2), abs(m1 - m2)), l1 + l2 + 1):  # if ld<abs(m1-m2) then P=0
                            A += a5_hash_table[idx_ab5]*sph[ld]*legendre[ld][abs(m1 - m2)] 
                            B += b5_hash_table[idx_ab5]*sph[ld]*legendre[ld][abs(m1 - m2)] 
                            idx_ab5 += 1
                            
                        A, B = eimph_val * A, eimph_val * B
                        for tau1 in range(2):
                            n1 = flds.multi_to_single_index(tau1, l1, m1, lmax1, mmax1)
                            for tau2 in range(2):
                                n2 = flds.multi_to_single_index(tau2, l2, m2, lmax2, mmax2)
                                if tau1 == tau2:
                                    w[n1, n2] = A
                                else:
                                    w[n1, n2] = B
    
        return w



##############################################################################
#  Make  hash table of ab5 and Legendre for a particle order pair 
#  This is for 2D coupling (Cython uses the GIL)
##############################################################################
# """
# This hash table is particularly good when all particles have the same order. 
# """

try: 
    from smuthi.utility.cython.cython_speedups import _ab5_coefficient_and_legendre_hash_table
    @memo.Memoize
    def ab5_coefficient_and_legendre_hash_table(lmax1, lmax2, mmax1, mmax2):
        return _ab5_coefficient_and_legendre_hash_table(int(lmax1), int(lmax2), int(mmax1), int(mmax2))
except:
    sys.stdout.write(
"""
Cython acceleration could not be loaded.
Falling back on Python equivalents... 
"""
        )
    sys.stdout.flush()
    @memo.Memoize
    def ab5_coefficient_and_legendre_hash_table(lmax1, lmax2, mmax1, mmax2):
        r"""Creates a hash table of the elements
        :math:`a5(l_1,m_1,l_2,m_2,l_d)*P_{l_d}^{m_1-m_2}(0)` and :math:`a5(l_1,m_1,l_2,m_2,l_d)*P_{l_d}^{m_1-m_2}(0)`
        found in appendix B of [Egel 2018 diss],where a5 and b5 are the coefficients used in the evaluation of the SVWF translation
        operator and :math:`P_l^m(\cos\theta)` are the normalized associated Legendre functions. This hash table is usefull in 
        reducing computation time when calculating the coupling between two particles that exist in the same layer 
        and have the same z coordinate.


        Args:
            lmax1 (int):           Largest polar quantum number of the recieving particle
            lmax2  (int):          Largest polar quantum number  of the emitting particle
            mmax1  (int):          Largest azimuthal quantum number of the recieving particle
            mmax2  (int):          Largest azimuthal quantum number of the emitting particle

    
        Returns:
            a5leg_array (ndarray):          hash table of the elements in :math:`A` not dependent on :math:`/phi` or :math:`kd`
            b5leg_array (ndarray):          hash table of the elements in :math:`B` not dependent on :math:`/phi` or :math:`kd`          

        """
        
        a5leg_array = []
        b5leg_array = []
        
        legendre_2D = sma.legendre_normalized(0, 1, lmax1+lmax2+1)[0][:,:,0]
        for im1, m1 in enumerate(range(-mmax1, mmax1 + 1)):
            for im2, m2 in enumerate(range(-mmax2, mmax2 + 1)):
                for il1, l1 in enumerate(range(max(1, abs(m1)), lmax1 + 1)):
                    for il2, l2 in enumerate(range(max(1, abs(m2)), lmax2 + 1)):
                        for ild, ld in enumerate(range(max(abs(l1 - l2), abs(m1 - m2)), l1 + l2 + 1)): 
                            
                            a5, b5 = trf.ab5_coefficients(l2, m2, l1, m1, ld)
                            leg =  legendre_2D[ld][abs(m1 - m2)]
                           
                            a5leg = a5*leg
                            b5leg = b5*leg
                            a5leg_array.append(a5leg)
                            b5leg_array.append(b5leg)
                          
    
        return a5leg_array, b5leg_array





##############################################################################
#  2D Direct coupling block using  hash table (Cython releases the Gil)
##############################################################################

try:
    from smuthi.utility.cython.cython_speedups import svwf_translate_2D_from_hash_table
except:  
    def svwf_translate_2D_from_hash_table(blocksize1,blocksize2,
                                    w,sph,k,dx, dy, dz,
                                    lmax1, lmax2, mmax1, mmax2,
                                    a5leg_hash_table,b5leg_hash_table,
                                    kind, threaded):

        r"""Subroutine to calculate the direct coupling between two particles 
            that are in the same layer and have the same z coordinate. This 
            subroutine is called internally by direct_coupling_block. 
            If Cython is enabled then this subroutine is Cython accelerated.
            Otherwise, the Python equivalent subroutine is used.
    
    
        Args:
            blocksize1 (int):           Number of columns in the direct coupling block
            blocksize2 (int):           Number of rows in the direct coupling block
            w  (ndarray):               Zero initialized direct coupling matrix block as numpy array
            sph  (ndarray):             Hankel or Bessel function matrix as numpy array
            k (complex):                Wavenumber in the shared media
            dx (float):                 x-coordinate of the distance between the two particles
            dy  (float):                y-coordinate of the distance between the two particles
            dz  (float):                z-coordinate of the distance between the two particles
            lmax1 (int):           Largest polar quantum number of the recieving particle
            lmax2  (int):          Largest polar quantum number  of the emitting particle
            mmax1  (int):          Largest azimuthal quantum number of the recieving particle
            mmax2  (int):          Largest azimuthal quantum number of the emitting particle
            a5leg_array (ndarray):          Hash table of the elements in :math:`A` not dependent on :math:`/phi` or :math:`kd`
            b5leg_array (ndarray):          Hash table of the elements in :math:`B` not dependent on :math:`/phi` or :math:`kd`          
            threaded (bool):          Flag to enable multithreading (valid only for Cython where the Gil can be released). Currently hard coded to False.  
            
        Returns:
            w  (ndarray):          Direct coupling matrix block as numpy array.
        """
    
    
        # calculate distance vector
        d = np.sqrt(dx**2 + dy**2 + dz**2)
        phi = np.arctan2(dy, dx)
        
        if kind == 0: # 0 = outgoing to incoming wave, 1 = outgoing to outgoing wave. 
        #NOTE: an long flag is used instead of a string because Cython 
        # has troublesome conventions about datatyping strings from python v2 to v3
        # See more on this at https://cython.readthedocs.io/en/latest/src/tutorial/strings.html
            for n in range(lmax1 + lmax2 + 1):
                sph[n] = sma.spherical_hankel(n, k*d) 
        else:
            for n in range(lmax1 + lmax2 + 1):
                sph[n] = sma.spherical_bessel(n, k*d) 
                
        # the particle coupling operator is the transpose of the SVWF translation operator
        # therefore, (l1,m1) and (l2,m2) are interchanged:
        idx_ab5leg = 0
        for m1 in range(-mmax1, mmax1 + 1):
            for m2 in range(-mmax2, mmax2 + 1):
                eimph_val = np.exp(1j * (m2 - m1) * phi)
                for l1 in range(max(1, abs(m1)), lmax1 + 1):
                    for l2 in range(max(1, abs(m2)), lmax2 + 1):
                        A, B = complex(0), complex(0)
                        for ld in range(max(abs(l1 - l2), abs(m1 - m2)), l1 + l2 + 1):  # if ld<abs(m1-m2) then P=0
    
                            A += a5leg_hash_table[idx_ab5leg]*sph[ld] 
                            B += b5leg_hash_table[idx_ab5leg]*sph[ld]
                            idx_ab5leg += 1
                            
                        A, B = eimph_val * A, eimph_val * B
                        for tau1 in range(2):
                            n1 = flds.multi_to_single_index(tau1, l1, m1, lmax1, mmax1)
                            for tau2 in range(2):
                                n2 = flds.multi_to_single_index(tau2, l2, m2, lmax2, mmax2)
                                if tau1 == tau2:
                                    w[n1, n2] = A
                                else:
                                    w[n1, n2] = B
    
        return w
    
    


"""This module contains CUDA source code for the preparation of coupling 
matrix lookups."""


# This cuda kernel is used for the calculation of volume lookup tables.
volume_lookup_assembly_code = """
    #define BLOCKSIZE %i
    #define RHO_ARRAY_LENGTH %i
    #define Z_ARRAY_LENGTH %i
    #define K_ARRAY_LENGTH %i
    
    __global__ void helper(const float *re_bes_jac, const float *im_bes_jac, const float *re_belbee, 
                            const float *im_belbee, const float *re_d_kappa, const float *im_d_kappa, 
                            float *re_result, float  *im_result)
    {
        unsigned int i = blockIdx.x * blockDim.x + threadIdx.x;
        if(i >= RHO_ARRAY_LENGTH * Z_ARRAY_LENGTH) return;
        
        unsigned int i_rho = i / Z_ARRAY_LENGTH;
        unsigned int i_z = i %% Z_ARRAY_LENGTH; 
        
        float re_res = 0.0;
        float im_res = 0.0;
        
        int i_kr = i_rho * K_ARRAY_LENGTH;
        int i_kz = i_z * K_ARRAY_LENGTH;

        float re_integrand_kp1 = re_bes_jac[i_kr] * re_belbee[i_kz] - im_bes_jac[i_kr] * im_belbee[i_kz];
        float im_integrand_kp1 = re_bes_jac[i_kr] * im_belbee[i_kz] + im_bes_jac[i_kr] * re_belbee[i_kz];

        for (int i_k=0; i_k<(K_ARRAY_LENGTH-1); i_k++)
        {
            i_kr = i_rho * K_ARRAY_LENGTH + i_k;
            i_kz = i_z * K_ARRAY_LENGTH + i_k;

            float re_integrand = re_integrand_kp1;
            float im_integrand = im_integrand_kp1;

            re_integrand_kp1 = re_bes_jac[i_kr+1] * re_belbee[i_kz+1] - im_bes_jac[i_kr+1] * im_belbee[i_kz+1];
            im_integrand_kp1 = re_bes_jac[i_kr+1] * im_belbee[i_kz+1] + im_bes_jac[i_kr+1] * re_belbee[i_kz+1];
            
            float re_sint = re_integrand + re_integrand_kp1;
            float im_sint = im_integrand + im_integrand_kp1;
            
            re_res += 0.5 * (re_sint * re_d_kappa[i_k] - im_sint * im_d_kappa[i_k]);
            im_res += 0.5 * (re_sint * im_d_kappa[i_k] + im_sint * re_d_kappa[i_k]);
        }
        
        re_result[i] = re_res;
        im_result[i] = im_res;
        
    }"""

# This cuda kernel is used for the calculation of radial lookup tables.
radial_lookup_assembly_code = """
    #define BLOCKSIZE %i
    #define RHO_ARRAY_LENGTH %i
    #define K_ARRAY_LENGTH %i
    
    __global__ void helper(const float *re_bes_jac, const float *im_bes_jac, const float *re_belbee, 
                            const float *im_belbee, const float *re_d_kappa, const float *im_d_kappa, 
                            float *re_result, float  *im_result)
    {
        unsigned int i = blockIdx.x * blockDim.x + threadIdx.x;
        if(i >= RHO_ARRAY_LENGTH) return;
        
        float re_res = 0.0;
        float im_res = 0.0;
        
        int i_kr = i * K_ARRAY_LENGTH;

        float re_integrand_kp1 = re_bes_jac[i_kr] * re_belbee[0] - im_bes_jac[i_kr] * im_belbee[0];
        float im_integrand_kp1 = re_bes_jac[i_kr] * im_belbee[0] + im_bes_jac[i_kr] * re_belbee[0];

        for (int i_k=0; i_k<(K_ARRAY_LENGTH-1); i_k++)
        {
            i_kr = i * K_ARRAY_LENGTH + i_k;
            
            float re_integrand = re_integrand_kp1;
            float im_integrand = im_integrand_kp1;

            re_integrand_kp1 = re_bes_jac[i_kr+1] * re_belbee[i_k+1] - im_bes_jac[i_kr+1] * im_belbee[i_k+1];
            im_integrand_kp1 = re_bes_jac[i_kr+1] * im_belbee[i_k+1] + im_bes_jac[i_kr+1] * re_belbee[i_k+1];
            
            float re_sint = re_integrand + re_integrand_kp1;
            float im_sint = im_integrand + im_integrand_kp1;
            
            re_res += 0.5 * (re_sint * re_d_kappa[i_k] - im_sint * im_d_kappa[i_k]);
            im_res += 0.5 * (re_sint * im_d_kappa[i_k] + im_sint * re_d_kappa[i_k]);
        }
        
        re_result[i] = re_res;
        im_result[i] = im_res;
        
    }"""
	
	
# This cuda kernel is used for the calculation of the direct radial lookup table
# based on the Plane Vector Wave Function coupling.
radial_lookup_assembly_code_PVWF = """
    #define BLOCKSIZE %i
    #define RHO_ARRAY_LENGTH %i
    #define K_ARRAY_LENGTH %i
    
    __global__ void helper(const float *re_bes_jac, const float *im_bes_jac, const float *re_bebe, 
                            const float *im_bebe, const float *re_d_kappa, const float *im_d_kappa, 
                            float *re_result, float  *im_result)
    {
        unsigned int i = blockIdx.x * blockDim.x + threadIdx.x;
        if(i >= RHO_ARRAY_LENGTH) return;
        
        float re_res = 0.0;
        float im_res = 0.0;
        
        int i_kr = i * K_ARRAY_LENGTH;

        float re_integrand_kp1 = re_bes_jac[i_kr] * re_bebe[i_kr] - im_bes_jac[i_kr] * im_bebe[i_kr];
        float im_integrand_kp1 = re_bes_jac[i_kr] * im_bebe[i_kr] + im_bes_jac[i_kr] * re_bebe[i_kr];

        for (int i_k=0; i_k<(K_ARRAY_LENGTH-1); i_k++)
        {
            i_kr = i * K_ARRAY_LENGTH + i_k;
            
            float re_integrand = re_integrand_kp1;
            float im_integrand = im_integrand_kp1;

            re_integrand_kp1 = re_bes_jac[i_kr+1] * re_bebe[i_kr+1] - im_bes_jac[i_kr+1] * im_bebe[i_kr+1];
            im_integrand_kp1 = re_bes_jac[i_kr+1] * im_bebe[i_kr+1] + im_bes_jac[i_kr+1] * re_bebe[i_kr+1];
            
            float re_sint = re_integrand + re_integrand_kp1;
            float im_sint = im_integrand + im_integrand_kp1;
            
            re_res += 0.5 * (re_sint * re_d_kappa[i_k] - im_sint * im_d_kappa[i_k]);
            im_res += 0.5 * (re_sint * im_d_kappa[i_k] + im_sint * re_d_kappa[i_k]);
        }
        
        re_result[i] = re_res;
        im_result[i] = im_res;
        
    }"""


"""
Routines for multiple scattering. Some modules contain functions to 
explicitly compute the coupling matrix entries. Others contains functions for 
the preparation of lookup tables that are used to approximate the coupling 
matrices by interoplation. The Cython folder has functions to calculate the 
direct coupling block between two particles in the same layer using fast C
routines.
"""



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Tue Apr  5 10:46:08 2022

This code is used to compile the cython .c file into machine code of the users
specific operating system. This method is the prefered method by Cython for 
distributing Cython code. 
https://cython.readthedocs.io/en/latest/src/userguide/source_files_and_compilation.html

The user should use this method to compile the existing .c file that will be 
distributed with smuthi. In this way, the user does not need to have Cython 
installed. They can compile using Python's setuptools module. 


Note: It is possible to combine with file with the compile_direct_coupling_block_in_c_from_pyx.py 
file using a "try" statement and adding an additional command line argument during 
compile time (See Cython link). But, I am unsure how the code will be compiled during user download 
of Smuthi. So, I opt to use the easiest code and have the other file exist only
for users who know what they are doing..


@author: parkerwray
"""


from setuptools import Extension, setup
import numpy

setup(
    ext_modules = [Extension("cython_speedups",
			 ["cython_speedups.c"],       
			 extra_compile_args=['-fopenmp'],
        		 extra_link_args=['-fopenmp'],
        		 include_dirs=[numpy.get_include()])]
)


"""
To compile direct_coupling_block_in_c.pyx, do the following:
    1) conda env list
    2) conda activate [env path]
        - Run python scripts with this env.
    3) python compile_direct_coupling_block_in_c_from_c.py build_ext --inplace
        - compile the c code using python to direct the compilation process.

"""



"""Functions to generate plots and animations."""

# -*- coding: utf-8 -*-
import numpy as np
import smuthi.postprocessing.scattered_field as sf
import smuthi.postprocessing.internal_field as intf
import smuthi.postprocessing.far_field as ff
import matplotlib.pyplot as plt
from matplotlib.patches import Circle, Ellipse, Rectangle
from matplotlib.colors import Normalize, LogNorm, SymLogNorm
from itertools import cycle
import tempfile
import shutil
#import imageio
import os
import warnings
import sys
from tqdm import tqdm


def plot_layer_interfaces(dim1min, dim1max, layer_system):
    """Add lines to plot to display layer system interfaces

    Args:
        dim1min (float):    From what x-value plot line
        dim1max (float):    To what x-value plot line
        layer_system (smuthi.layers.LayerSystem):   Stratified medium
    """
    for il in range(1, layer_system.number_of_layers()):
        plt.plot([dim1min, dim1max], [layer_system.reference_z(il), layer_system.reference_z(il)], 'g')


def plot_particles(xmin, xmax, ymin, ymax, zmin, zmax, particle_list,
                   draw_circumscribing_sphere, fill_particle=True):
    """Add circles, ellipses and rectangles to plot to display spheres, spheroids and cylinders.

    Args:
        xmin (float):   Minimal x-value of plot
        xmax (float):   Maximal x-value of plot
        ymin (float):   Minimal y-value of plot
        ymax (float):   Maximal y-value of plot
        zmin (float):   Minimal z-value of plot
        zmax (float):   Maximal z-value of plot
        particle_list (list): List of smuthi.particles.Particle objects
        draw_circumscribing_sphere (bool): If true (default), draw a circle indicating the circumscribing sphere of
                                           particles.
        fill_particle (bool):              If true, draw opaque particles.
    """

    ax = plt.gca()

    if fill_particle:
        particle_face_color = 'k'
    else:
        particle_face_color = 'none'

    if xmin == xmax:
        plane_coord = 0
        draw_coord = [1, 2]
    elif ymin == ymax:
        plane_coord = 1
        draw_coord = [0, 2]
    elif zmin == zmax:
        plane_coord = 2
        draw_coord = [0, 1]
    else:
        raise ValueError('Field points must define a plane')

    for particle in particle_list:
        pos = particle.position
        dis = abs((xmin, ymin, zmin)[plane_coord] - pos[plane_coord])

        if dis > particle.circumscribing_sphere_radius():
            continue

        circradius = np.sqrt(particle.circumscribing_sphere_radius()**2 - dis**2)
        if type(particle).__name__ == 'Sphere':
            ax.add_patch(Circle((pos[draw_coord[0]], pos[draw_coord[1]]), circradius,
                         facecolor=particle_face_color, edgecolor='k'))
        else:
            if not particle.euler_angles == [0, 0, 0]:
                warnings.warn("Drawing rotated particles currently not supported - drawing black disc with size"
                              + " based on the circumscribing sphere instead")
                ax.add_patch(Circle((pos[draw_coord[0]], pos[draw_coord[1]]), circradius, facecolor='k', edgecolor='k'))
                ax.text(pos[draw_coord[0]], pos[draw_coord[1]], 'rotated ' + type(particle).__name__,
                        verticalalignment='center', horizontalalignment='center', color='blue', fontsize=5)
            else:
                if draw_circumscribing_sphere:
                    ax.add_patch(Circle((pos[draw_coord[0]], pos[draw_coord[1]]), circradius,
                                 linestyle='dashed', facecolor='none', edgecolor='k'))

                if type(particle).__name__ == 'Spheroid':
                    a = particle.semi_axis_a
                    c = particle.semi_axis_c
                    if plane_coord in [0, 1]:
                        w = a * np.sqrt(1 - (dis/a)**2) if dis < a else 0
                        h = c * np.sqrt(1 - (dis/a)**2) if dis < a else 0
                    else:
                        w = a * np.sqrt(1 - (dis/c)**2) if dis < c else 0
                        h = w
                    ax.add_patch(Ellipse(xy=(pos[draw_coord[0]], pos[draw_coord[1]]), width=2*w, height=2*h,
                                         facecolor='w', edgecolor='k'))

                elif type(particle).__name__ == 'FiniteCylinder':
                    r = particle.cylinder_radius
                    h = particle.cylinder_height
                    if plane_coord in [0, 1]:
                        w = np.sqrt(r**2 - dis**2) if dis < r else 0
                        h *= np.sign(w)
                        ax.add_patch(Rectangle((pos[draw_coord[0]]-w, pos[draw_coord[1]]-h/2),
                                                2*w, h, facecolor='w', edgecolor='k'))
                    elif np.logical_and(plane_coord == 2, dis <= h/2):
                        ax.add_patch(Circle((pos[draw_coord[0]], pos[draw_coord[1]]), r,
                                            facecolor='w', edgecolor='k'))


def compute_near_field(simulation=None, X=None, Y=None, Z=None, type='scatt', chunksize=None,
                       k_parallel='default', azimuthal_angles='default', angular_resolution=None):
    """Compute a certain component of the electric near field"""
    X, Y, Z = np.atleast_1d(X).T, np.atleast_1d(Y).T, np.atleast_1d(Z).T
    e_x, e_y, e_z = (np.zeros_like(X, dtype=np.complex128) for i in range(3))

    if chunksize is None: # pick chunksize dynamically between 4096, 8192 or 16384
        chunksize = min(2**16, 2**np.floor(np.log2(X.size / 10))).astype(int) if X.size > 2**12 else 2**12

    if 'sca' in type:
        min_laynum = simulation.layer_system.layer_number(Z.min())
        max_laynum = simulation.layer_system.layer_number(Z.max())
        layer_numbers = [i for i in range(min_laynum, max_laynum + 1)]
        scat_fld_exp = sf.scattered_field_piecewise_expansion(simulation.initial_field.vacuum_wavelength,
                                                              simulation.particle_list,
                                                              simulation.layer_system,
                                                              k_parallel,
                                                              azimuthal_angles,
                                                              angular_resolution,
                                                              layer_numbers)
    if 'int' in type:
        int_fld_exp = intf.internal_field_piecewise_expansion(simulation.initial_field.vacuum_wavelength, simulation.particle_list,
                                                              simulation.layer_system)

    descr = 'Compute '+type+' near-field'

    nchunks = np.ceil(X.size / chunksize).astype(int)
    e_x, e_y, e_z = ([None] * nchunks for i in range(3))

    Xc, Yc, Zc = np.array_split(X, nchunks), np.array_split(Y, nchunks), np.array_split(Z, nchunks)

    for c in tqdm(range(len(Xc)), desc=descr.ljust(26), file=sys.stdout,
                                  bar_format='{l_bar}{bar}| elapsed: {elapsed} ' 'remaining: {remaining}'):
        xarr, yarr, zarr = Xc[c], Yc[c], Zc[c]
        if 'sca' in type:
            e_x[c], e_y[c], e_z[c] = scat_fld_exp.electric_field(xarr, yarr, zarr)
        elif 'ini' in type:
            e_x[c], e_y[c], e_z[c] = simulation.initial_field.electric_field(xarr, yarr, zarr, simulation.layer_system)
        elif 'int' in type:
            e_x[c], e_y[c], e_z[c] = int_fld_exp.electric_field(xarr, yarr, zarr)

    return np.concatenate(e_x), np.concatenate(e_y), np.concatenate(e_z)


def show_near_field(simulation=None, quantities_to_plot=None,
                    show_plots=True, show_opts=None, save_plots=False, save_opts=None,
                    save_data=False, data_format='hdf5', outputdir='.',
                    xmin=0, xmax=0, ymin=0, ymax=0, zmin=0, zmax=0,
                    resolution_step=25, k_parallel='default', azimuthal_angles='default', angular_resolution=None,
                    draw_circumscribing_sphere=True, show_internal_field=False):
    """Plot the electric near field along a plane. To plot along the xy-plane, specify zmin=zmax and so on.

    Args:
        simulation (smuthi.simulation.Simulation):  Simulation object
        quantities_to_plot:     List of strings that specify what to plot. Select from 'E_x', 'E_y', 'E_z', 'norm(E)'
                                The list may contain one or more of the following strings:

                                'E_x'           real part of x-component of complex total electric field
                                'E_y'           real part of y-component of complex total electric field
                                'E_z'           real part of z-component of complex total electric field
                                'norm(E)'       norm of complex total electric field

                                'E_scat_x'      real part of x-component of complex scattered electric field
                                'E_scat_y'      real part of y-component of complex scattered electric field
                                'E_scat_z'      real part of z-component of complex scattered electric field
                                'norm(E_scat)'  norm of complex scattered electric field

                                'E_init_x'      real part of x-component of complex initial electric field
                                'E_init_y'      real part of y-component of complex initial electric field
                                'E_init_z'      real part of z-component of complex initial electric field
                                'norm(E_init)'  norm of complex initial electric field
        show_plots (logical):   If True, plots are shown
        show_opts (dict list):  List of dictionaries containing options to be passed to imshow for plotting.
                                For each entry in quantities_to_plot, all show_opts dictionaries will be applied.
                                If save_plots=True, a 1:1 correspondence between show_opts and save_opts dictionaries
                                is assumed. For simplicity, one can also provide a single show_opts entry that will
                                be applied to all save_opts.
                                The following keys are made available (see matplotlib.pyplot.imshow documentation):
                                'cmap'          defaults to 'inferno' for norm quantities and 'RdYlBu' otherwise
                                'norm'          (None). If a norm is provided, its vmin and vmax take precedence
                                'aspect'        ('equal')
                                'interpolation' (None), also available: bilinear, bicubic, spline16, quadric, ...
                                'alpha'         (None)
                                'vmin'          (None), will be set to 0 for norm quantities and -vmax otherwise
                                'vmax'          initialized with the max of the quantity to plot
                                'origin'        ('lower')
                                'extent'        calculated automatically based on plotting coordinate limits
                                An optional extra key called 'label' of type string is shown in the plot title
                                and appended to the associated file if save_plots is True
                                Finally, an optional 'figsize' key is available to set the width and height of
                                the figure window (see matplotlib.pyplot.figure documentation)
        save_plots (logical):   If True, plots are exported to file.
        save_opts (dict list):  List of dictionaries containing options to be passed to savefig.
                                For each entry in quantities_to_plot, all save_opts dictionaries will be applied.
                                A 1:1 correspondence between save_opts and show_opts dictionaries is assumed. For
                                simplicity, one can also provide a single save_opts entry that will be applied to
                                all show_opts.
                                The following keys are made available (see matplotlib.pyplot.savefig documentation):
                                'dpi'           (None)
                                'orientation'   (None)
                                'format'        ('png'), also available: eps, jpeg, jpg, pdf, ps, svg, tif, tiff ...
                                'transparent'   (False)
                                'bbox_inches'   ('tight')
                                'pad_inches'    (0.1)
                                Passing 'gif' as one of the format values will result in an animation if the
                                quantity to plot is of non-norm type
        save_data (logical):    If True, raw data are exported to file
        data_format (str):      Output data format string, 'hdf5' and 'ascii' formats are available
        outputdir (str):        Path to directory where to save the export files
        xmin (float):           Plot from that x (length unit)
        xmax (float):           Plot up to that x (length unit)
        ymin (float):           Plot from that y (length unit)
        ymax (float):           Plot up to that y (length unit)
        zmin (float):           Plot from that z (length unit)
        zmax (float):           Plot up to that z (length unit)
        resolution_step (float):     Compute the field with that spatial resolution (length unit,
                                     distance between computed points), can be a tuple for [resx, resy, resz]
        k_parallel (numpy.ndarray or str):      in-plane wavenumbers for the plane wave expansion
                                                if 'default', use smuthi.fields.default_Sommerfeld_k_parallel_array
        azimuthal_angles (numpy.ndarray or str):azimuthal angles for the plane wave expansion
                                                if 'default', use smuthi.fields.default_azimuthal_angles
        angular_resolution (float):             If provided, angular arrays are generated with this angular
                                                resolution over the default angular range
        draw_circumscribing_sphere (bool):      If true (default), draw a circle indicating the circumscribing
                                                sphere of particles.
        show_internal_field (bool):             If true, compute also the field inside the particles (only for spheres)
    """

    if (not os.path.exists(outputdir)) and (save_plots or save_data):
        os.makedirs(outputdir)

    if quantities_to_plot is None:
        quantities_to_plot = ['norm(E)']
    if show_opts is None:
        show_opts = [{'interpolation':'none'}]
    if save_opts is None:
        save_opts = [{'format':'png'}]

    res = np.resize(resolution_step,3)
    x = np.arange(xmin, xmax + res[0]/2, res[0])
    y = np.arange(ymin, ymax + res[1]/2, res[1])
    z = np.arange(zmin, zmax + res[2]/2, res[2])
    xarr, yarr, zarr = np.squeeze(np.meshgrid(x, y, z, indexing='ij'))

    if xmin == xmax:
        dim1vec, dim2vec = y, z
        step1, step2 = res[1]/2, res[2]/2 # these are for proper registration of particles and layers to the pixel grid
        titlestr = '$x$ = {} '.format(xmin) + simulation.length_unit
        dim1name = '$y$ (' + simulation.length_unit + ')'
        dim2name = '$z$ (' + simulation.length_unit + ')'
    elif ymin == ymax:
        dim1vec, dim2vec = x, z
        step1, step2 = res[0]/2, res[2]/2
        titlestr = '$y$ = {} '.format(ymin) + simulation.length_unit
        dim1name = '$x$ (' + simulation.length_unit + ')'
        dim2name = '$z$ (' + simulation.length_unit + ')'
    elif zmin == zmax:
        dim1vec, dim2vec = x, y
        step1, step2 = res[0]/2, res[1]/2
        titlestr = '$z$ = {} '.format(zmin) + simulation.length_unit
        dim1name = '$x$ (' + simulation.length_unit + ')'
        dim2name = '$y$ (' + simulation.length_unit + ')'

    sys.stdout.write("Evaluate fields ...\n")
    sys.stdout.flush()
    e_x_scat, e_y_scat, e_z_scat = compute_near_field(simulation=simulation, X=xarr, Y=yarr, Z=zarr, type='scatt.')
    e_x_init, e_y_init, e_z_init = compute_near_field(simulation=simulation, X=xarr, Y=yarr, Z=zarr, type='initl.')

    if show_internal_field:
        e_x_int, e_y_int, e_z_int = compute_near_field(simulation=simulation, X=xarr, Y=yarr, Z=zarr, type='intrn.')

    sys.stdout.write("Generate plots ...\n")
    sys.stdout.flush()

    if not show_plots:
        import matplotlib
        default_backend = matplotlib.get_backend()
        matplotlib.use('Agg') # a non-gui backend to avoid opening figure windows

    for quantity in quantities_to_plot:

        for show_opt, save_opt in zip(cycle(show_opts), save_opts) if len(show_opts) < len(save_opts) else zip(show_opts, cycle(save_opts)):

            try:
                filename = 'E'
                label_str = show_opt.get('label','')

                if 'scat' in quantity:
                    e_x, e_y, e_z = e_x_scat, e_y_scat, e_z_scat
                    field_type_string = 'scat'
                elif 'init' in quantity:
                    e_x, e_y, e_z = e_x_init, e_y_init, e_z_init
                    field_type_string = 'init'
                elif 'intern' in quantity:
                    if not show_internal_field:
                        raise Exception("show_internal_field flag needs to be set to true!")
                    e_x, e_y, e_z = e_x_int, e_y_int, e_z_int
                    field_type_string = 'inter'
                else:
                    e_x, e_y, e_z = e_x_scat + e_x_init, e_y_scat + e_y_init, e_z_scat + e_z_init
                    if show_internal_field:
                        e_x, e_y, e_z = e_x + e_x_int, e_y + e_y_int, e_z + e_z_int
                    field_type_string = 'tot'

                fig = plt.figure(figsize=show_opt.get('figsize',[6.4, 4.8]))
                if not zmin == zmax:
                    plot_layer_interfaces(dim1vec[0], dim1vec[-1], simulation.layer_system)
                plot_particles(xmin, xmax, ymin, ymax, zmin, zmax, simulation.particle_list,
                               draw_circumscribing_sphere, not show_internal_field)
                if 'norm' in quantity:
                    e = np.sqrt(abs(e_x)**2 + abs(e_y)**2 + abs(e_z)**2)
                    vmax = np.abs(e).max()
                    color_norm = show_opt.get('norm', Normalize(vmin=show_opt.get('vmin',0), vmax=show_opt.get('vmax',vmax)))
                    plt.imshow(e,
                               alpha=show_opt.get('alpha'), norm=color_norm,
                               cmap=show_opt.get('cmap','inferno'), origin=show_opt.get('origin','lower'),
                               interpolation=show_opt.get('interpolation','none'),
                               extent=show_opt.get('extent', [dim1vec.min()-step1, dim1vec.max()+step1,
                                                              dim2vec.min()-step2, dim2vec.max()+step2]))
                    plt_title = '$|' + filename + '^{' + field_type_string + '}|$ at ' + titlestr \
                                + ('' if label_str == '' else ' (' + label_str + ')')
                    filename = 'norm_' + filename
                    plt.title(plt_title)
                else:
                    if '_x' in quantity:
                        e = e_x
                        filename = filename + '_x'
                    elif '_y' in quantity:
                        e = e_y
                        filename = filename + '_y'
                    elif '_z' in quantity:
                        e = e_z
                        filename = filename + '_z'
                    else:
                        print('Quantity:', quantity)
                        raise ValueError('field component not specified')

                    vmax = np.abs(e).max()
                    color_norm = show_opt.get('norm', Normalize(vmin=show_opt.get('vmin',-vmax), vmax=show_opt.get('vmax',vmax)))
                    plt.imshow(e.real,
                               alpha=show_opt.get('alpha'), norm=color_norm,
                               cmap=show_opt.get('cmap','RdYlBu'), origin=show_opt.get('origin','lower'),
                               interpolation=show_opt.get('interpolation','none'), aspect=show_opt.get('aspect','equal'),
                               extent=show_opt.get('extent', [dim1vec.min()-step1, dim1vec.max()+step1,
                                                              dim2vec.min()-step2, dim2vec.max()+step2]))
                    plt_title = '$ ' + filename + '^{' + field_type_string + '}$' + ' at ' + titlestr \
                                + ('' if label_str == '' else ' (' + label_str + ')')
                    plt.title(plt_title)

                plt.colorbar()
                plt.xlabel(dim1name)
                plt.ylabel(dim2name)

                label_str = '' if label_str == '' else '_' + label_str # prepend underscore
                filename = filename + '_' + field_type_string + label_str
                export_filename = outputdir + '/' + filename
                if save_plots:
                    if save_opt.get('format') == 'gif':
                        if 'norm' in quantity:
                            plt.close(fig)
                            continue # it seems like savefig does not accept 'gif' as a format, so we have to continue explicitly here
                        tempdir = tempfile.mkdtemp()
                        images = []
                        for i_t, t in enumerate(np.linspace(0, 1, 20, endpoint=False)):
                            tempfig = plt.figure(figsize=show_opt.get('figsize',[6.4, 4.8]))
                            if not zmin == zmax:
                                plot_layer_interfaces(dim1vec[0], dim1vec[-1], simulation.layer_system)
                            plot_particles(xmin, xmax, ymin, ymax, zmin, zmax, simulation.particle_list,
                                           draw_circumscribing_sphere, not show_internal_field)
                            e_t = e * np.exp(-1j * t * 2 * np.pi)
                            plt.imshow(e_t.real,
                                       alpha=show_opt.get('alpha'), norm=color_norm,
                                       cmap=show_opt.get('cmap','RdYlBu'), origin=show_opt.get('origin','lower'),
                                       interpolation=show_opt.get('interpolation','none'), aspect=show_opt.get('aspect','equal'),
                                       extent=show_opt.get('extent', [dim1vec.min()-step1, dim1vec.max()+step1,
                                                                      dim2vec.min()-step2, dim2vec.max()+step2]))
                            plt.title(plt_title)
                            plt.colorbar()
                            plt.xlabel(dim1name)
                            plt.ylabel(dim2name)

                            tempfig_filename = tempdir + '/temp_' + str(i_t) + '.png'
                            plt.savefig(tempfig_filename, dpi=save_opt.get('dpi'),
                                        orientation=save_opt.get('orientation','portrait'),
                                        transparent=save_opt.get('transparent',False),
                                        bbox_inches=save_opt.get('bbox_inches','tight'),
                                        pad_inches=save_opt.get('pad_inches',0.1))
                            plt.close(tempfig)
                            images.append(imageio.imread(tempfig_filename))
                        imageio.mimsave(export_filename + '.gif', images, duration=0.1)
                        shutil.rmtree(tempdir)
                    else:
                        file_ext = '.' + save_opt.get('format','png') # default to png if not specified
                        plt.savefig(export_filename + file_ext, dpi=save_opt.get('dpi'),
                                    orientation=save_opt.get('orientation','portrait'),
                                    transparent=save_opt.get('transparent',False),
                                    bbox_inches=save_opt.get('bbox_inches','tight'),
                                    pad_inches=save_opt.get('pad_inches',0.1))

                if not show_plots:
                    plt.close(fig)

            except Exception as e: # TODO: parse to understand if this is due to incompatible options or trying to plot non-2D data
                print("Skipping " + quantity + " for show_opt = ", show_opt, " and save_opt = ", save_opt)

    if not show_plots:
        matplotlib.use(default_backend) # now we can restore the original backend
    else:
        plt.show(block=False)

    if save_data:
        Ei = np.stack((e_x_init, e_y_init, e_z_init))
        Es = np.stack((e_x_scat, e_y_scat, e_z_scat))
        if show_internal_field:
            En = np.stack((e_x_int, e_y_int, e_z_int))

        if data_format.lower() == 'hdf5':
            import h5py
            # TODO: move metadata writing to auxiliary function
            # TODO: add metadata relative to particles and layer system ?
            if type(simulation.initial_field).__name__ == "DipoleSource":
                metadata = {'initial_field': type(simulation.initial_field).__name__,
                            'dipole_moment': simulation.initial_field.dipole_moment,
                            'dipole_position': simulation.initial_field.position,}
            elif type(simulation.initial_field).__name__ == "DipoleCollection":
                metadata = {'initial_field': type(simulation.initial_field).__name__,
                            'dipole_moments': [dip.dipole_moment for dip in simulation.initial_field.dipole_list],
                            'dipole_positions': [dip.position for dip in simulation.initial_field.dipole_list],}
            elif type(simulation.initial_field).__name__ == "PlaneWave":
                metadata = {'initial_field': type(simulation.initial_field).__name__,
                            'polar_angle': simulation.initial_field.polar_angle,
                            'azimuthal_angle': simulation.initial_field.azimuthal_angle,
                            'polarization': simulation.initial_field.polarization,}
            elif type(simulation.initial_field).__name__ == "GaussianBeam":
                metadata = {'initial_field': type(simulation.initial_field).__name__,
                            'polar_angle': simulation.initial_field.polar_angle,
                            'azimuthal_angle': simulation.initial_field.azimuthal_angle,
                            'polarization': simulation.initial_field.polarization,
                            'reference_point': simulation.initial_field.reference_point,
                            'beam_waist': simulation.initial_field.beam_waist,}

            with h5py.File(outputdir + '/data.hdf5', 'a') as f:
                # add attributes that are dependent on the initial field
                f.attrs.update(metadata)
                # additional attributes that are common to all simulations
                f.attrs['vacuum_wavelength'] = simulation.initial_field.vacuum_wavelength
                # write actual data
                g = f.require_group('near_field')
                g.create_dataset('x', data=x)
                g.create_dataset('y', data=y)
                g.create_dataset('z', data=z)
                g.create_dataset('init_electric_field', data=Ei, dtype='c16', compression="gzip")
                g.create_dataset('scat_electric_field', data=Es, dtype='c16', compression="gzip")
                if show_internal_field:
                    g.create_dataset('inte_electric_field', data=En, dtype='c16', compression="gzip")
        elif data_format.lower() == 'ascii':
            np.savetxt(outputdir + '/1st_dim_axis.dat', x, fmt='%g')
            np.savetxt(outputdir + '/2nd_dim_axis.dat', y, fmt='%g')
            np.savetxt(outputdir + '/3rd_dim_axis.dat', z, fmt='%g')
            fmt = list(np.repeat('%.15g %+.15gj', np.size(dim1vec)))
            np.savetxt(outputdir + '/e_init_x.dat', Ei[0,], fmt=fmt, delimiter=',')
            np.savetxt(outputdir + '/e_init_y.dat', Ei[1,], fmt=fmt, delimiter=',')
            np.savetxt(outputdir + '/e_init_z.dat', Ei[2,], fmt=fmt, delimiter=',')
            np.savetxt(outputdir + '/e_scat_x.dat', Es[0,], fmt=fmt, delimiter=',')
            np.savetxt(outputdir + '/e_scat_y.dat', Es[1,], fmt=fmt, delimiter=',')
            np.savetxt(outputdir + '/e_scat_z.dat', Es[2,], fmt=fmt, delimiter=',')
            if show_internal_field:
                np.savetxt(outputdir + '/e_int_x.dat', En[0,], fmt=fmt, delimiter=',')
                np.savetxt(outputdir + '/e_int_y.dat', En[1,], fmt=fmt, delimiter=',')
                np.savetxt(outputdir + '/e_int_z.dat', En[2,], fmt=fmt, delimiter=',')
        else:
            raise ValueError('Currently, only hdf5 or ascii output data formats are available')


def show_scattered_far_field(simulation, show_plots=True, show_opts=[{'label':'scattered_far_field'}],
                             save_plots=False, save_opts=None,
                             save_data=False, data_format='hdf5', outputdir='.',
                             flip_downward=True, split=True, log_scale=False,
                             polar_angles='default', azimuthal_angles='default', angular_resolution=None):
    """Display and export the scattered far field.

    Args:
        simulation (smuthi.simulation.Simulation):  Simulation object
        show_plots (bool):                          Display plots if True
        show_opts (dict list):  List of dictionaries containing options to be passed to pcolormesh for plotting.
                                If save_plots=True, a 1:1 correspondence between show_opts and save_opts dictionaries
                                is assumed. For simplicity, one can also provide a single show_opts entry that will
                                be applied to all save_opts.
                                The following keys are available (see matplotlib.pyplot.pcolormesh documentation):
                                'alpha'     (None)
                                'cmap'      ('inferno')
                                'norm'      (None), is set to matplotlib.colors.LogNorm() if log_scale is True
                                'vmin'      (None), applies only to 2D plots
                                'vmax'      (None), applies only to 2D plots
                                'shading'   ('nearest'), applies only to 2D plots. 'gouraud' is also available
                                'linewidth' (None), applies only to 1D plots
                                'linestyle' (None), applies only to 1D plots
                                'marker'    (None), applies only to 1D plots
                                An optional extra key called 'label' of type string is shown in the plot title
                                and appended to the associated file if save_plots is True
        save_plots (bool):      If True, plots are exported to file.
        save_opts (dict list):  List of dictionaries containing options to be passed to savefig.
                                A 1:1 correspondence between save_opts and show_opts dictionaries is assumed. For
                                simplicity, one can also provide a single save_opts entry that will be applied to
                                all show_opts.
                                The following keys are made available (see matplotlib.pyplot.savefig documentation):
                                'dpi'           (None)
                                'orientation'   (None)
                                'format'        ('png'), also available: eps, jpeg, jpg, pdf, ps, svg, tif, tiff ...
                                'transparent'   (False)
                                'bbox_inches'   ('tight')
                                'pad_inches'    (0.1)
        save_data (bool):       If True, raw data are exported to file
        data_format (str):      Output data format string, 'hdf5' and 'ascii' formats are available
        outputdir (str):                        Path to the directory where files are to be saved
        flip_downward (bool):                   If True, represent downward directions as 0-90 deg instead of 90-180
        split (bool):                           If True, show two different plots for upward and downward directions
        log_scale (bool):                       If True, set a logarithmic scale
        polar_angles (numpy.ndarray or str):    Polar angles values (radian).
                                                If 'default', use smuthi.fields.default_polar_angles
        azimuthal_angles (numpy.ndarray or str):Azimuthal angle values (radian).
                                                If 'default', use smuthi.fields.default_azimuthal_angles
        angular_resolution (float):             If provided, angular arrays are generated with this angular resolution
                                                over the default angular range
    """

    infld = simulation.initial_field
    plst = simulation.particle_list
    lsys = simulation.layer_system
    far_field = ff.scattered_far_field(vacuum_wavelength=infld.vacuum_wavelength,
                                       particle_list=plst,
                                       layer_system=lsys,
                                       polar_angles=polar_angles,
                                       azimuthal_angles=azimuthal_angles,
                                       angular_resolution=angular_resolution)

    [d.setdefault('label','scattered_far_field') for d in show_opts]

    show_far_field(far_field=far_field, save_plots=save_plots, save_opts=save_opts, show_plots=show_plots,
                   show_opts=show_opts, save_data=save_data, data_format=data_format, outputdir=outputdir,
                   flip_downward=flip_downward, split=split, log_scale=log_scale)


def show_total_far_field(simulation, show_plots=True, show_opts=[{'label':'total_far_field'}],
                         save_plots=False, save_opts=None,
                         save_data=False, data_format='hdf5', outputdir='.',
                         flip_downward=True, split=True, log_scale=False,
                         polar_angles='default', azimuthal_angles='default', angular_resolution=None):
    """Display and export the total far field. This function cannot be used if the inital field is a plane wave.

    Args:
        simulation (smuthi.simulation.Simulation):  Simulation object
        show_plots (bool):                          Display plots if True
        show_opts (dict list):  List of dictionaries containing options to be passed to pcolormesh for plotting.
                                If save_plots=True, a 1:1 correspondence between show_opts and save_opts dictionaries
                                is assumed. For simplicity, one can also provide a single show_opts entry that will
                                be applied to all save_opts.
                                The following keys are available (see matplotlib.pyplot.pcolormesh documentation):
                                'alpha'     (None)
                                'cmap'      ('inferno')
                                'norm'      (None), is set to matplotlib.colors.LogNorm() if log_scale is True
                                'vmin'      (None), applies only to 2D plots
                                'vmax'      (None), applies only to 2D plots
                                'shading'   ('nearest'), applies only to 2D plots. 'gouraud' is also available
                                'linewidth' (None), applies only to 1D plots
                                'linestyle' (None), applies only to 1D plots
                                'marker'    (None), applies only to 1D plots
                                An optional extra key called 'label' of type string is shown in the plot title
                                and appended to the associated file if save_plots is True
        save_plots (bool):      If True, plots are exported to file.
        save_opts (dict list):  List of dictionaries containing options to be passed to savefig.
                                A 1:1 correspondence between save_opts and show_opts dictionaries is assumed. For
                                simplicity, one can also provide a single save_opts entry that will be applied to
                                all show_opts.
                                The following keys are made available (see matplotlib.pyplot.savefig documentation):
                                'dpi'           (None)
                                'orientation'   (None)
                                'format'        ('png'), also available: eps, jpeg, jpg, pdf, ps, svg, tif, tiff ...
                                'transparent'   (False)
                                'bbox_inches'   ('tight')
                                'pad_inches'    (0.1)
        save_data (bool):       If True, raw data are exported to file
        data_format (str):      Output data format string, 'hdf5' and 'ascii' formats are available
        outputdir (str):                        Path to the directory where files are to be saved
        flip_downward (bool):                   If True, represent downward directions as 0-90 deg instead of 90-180
        split (bool):                           If True, show two different plots for upward and downward directions
        log_scale (bool):                       If True, set a logarithmic scale
        polar_angles (numpy.ndarray or str):    Polar angles values (radian).
                                                If 'default', use smuthi.fields.default_polar_angles
        azimuthal_angles (numpy.ndarray or str):Azimuthal angle values (radian).
                                                If 'default', use smuthi.fields.default_azimuthal_angles
        angular_resolution (float):             If provided, angular arrays are generated with this angular resolution
                                                over the default angular range
    """
    infld = simulation.initial_field
    plst = simulation.particle_list
    lsys = simulation.layer_system
    far_field, _, _ = ff.total_far_field(initial_field=infld,
                                         particle_list=plst,
                                         layer_system=lsys,
                                         polar_angles=polar_angles,
                                         azimuthal_angles=azimuthal_angles,
                                         angular_resolution=angular_resolution)

    [d.setdefault('label','total_far_field') for d in show_opts]

    show_far_field(far_field=far_field, save_plots=save_plots, save_opts=save_opts, show_plots=show_plots,
                   show_opts=show_opts, save_data=save_data, data_format=data_format, outputdir=outputdir,
                   flip_downward=flip_downward, split=split, log_scale=log_scale)


def show_scattering_cross_section(simulation, show_plots=True, show_opts=[{'label':'scattering_cross_section'}],
                                  save_plots=False, save_opts=None,
                                  save_data=False, data_format='hdf5', outputdir='.',
                                  flip_downward=True, split=True, log_scale=False,
                                  polar_angles='default', azimuthal_angles='default', angular_resolution=None):
    """Display and export the differential scattering cross section.

    Args:
        simulation (smuthi.simulation.Simulation):  Simulation object
        show_plots (bool):                          Display plots if True
        show_opts (dict list):  List of dictionaries containing options to be passed to pcolormesh for plotting.
                                If save_plots=True, a 1:1 correspondence between show_opts and save_opts dictionaries
                                is assumed. For simplicity, one can also provide a single show_opts entry that will
                                be applied to all save_opts.
                                The following keys are available (see matplotlib.pyplot.pcolormesh documentation):
                                'alpha'     (None)
                                'cmap'      ('inferno')
                                'norm'      (None), is set to matplotlib.colors.LogNorm() if log_scale is True
                                'vmin'      (None), applies only to 2D plots
                                'vmax'      (None), applies only to 2D plots
                                'shading'   ('nearest'), applies only to 2D plots. 'gouraud' is also available
                                'linewidth' (None), applies only to 1D plots
                                'linestyle' (None), applies only to 1D plots
                                'marker'    (None), applies only to 1D plots
                                An optional extra key called 'label' of type string is shown in the plot title
                                and appended to the associated file if save_plots is True
        save_plots (bool):      If True, plots are exported to file.
        save_opts (dict list):  List of dictionaries containing options to be passed to savefig.
                                A 1:1 correspondence between save_opts and show_opts dictionaries is assumed. For
                                simplicity, one can also provide a single save_opts entry that will be applied to
                                all show_opts.
                                The following keys are made available (see matplotlib.pyplot.savefig documentation):
                                'dpi'           (None)
                                'orientation'   (None)
                                'format'        ('png'), also available: eps, jpeg, jpg, pdf, ps, svg, tif, tiff ...
                                'transparent'   (False)
                                'bbox_inches'   ('tight')
                                'pad_inches'    (0.1)
        save_data (bool):       If True, raw data are exported to file
        data_format (str):      Output data format string, 'hdf5' and 'ascii' formats are available
        outputdir (str):                        Path to the directory where files are to be saved
        flip_downward (bool):                   If True, represent downward directions as 0-90 deg instead of 90-180
        split (bool):                           If True, show two different plots for upward and downward directions
        log_scale (bool):                       If True, set a logarithmic scale
        polar_angles (numpy.ndarray or str):    Polar angles values (radian).
                                                If 'default', use smuthi.fields.default_polar_angles
        azimuthal_angles (numpy.ndarray or str):Azimuthal angle values (radian).
                                                If 'default', use smuthi.fields.default_azimuthal_angles
        angular_resolution (float):             If provided, angular arrays are generated with this angular resolution
                                                over the default angular range
    """

    infld = simulation.initial_field
    plst = simulation.particle_list
    lsys = simulation.layer_system
    far_field = ff.scattering_cross_section(initial_field=infld,
                                            particle_list=plst,
                                            layer_system=lsys,
                                            polar_angles=polar_angles,
                                            azimuthal_angles=azimuthal_angles,
                                            angular_resolution=angular_resolution)

    [d.setdefault('label','scattering_cross_section') for d in show_opts]

    show_far_field(far_field=far_field, save_plots=save_plots, save_opts=save_opts, show_plots=show_plots,
                   show_opts=show_opts, save_data=save_data, data_format=data_format, outputdir=outputdir,
                   flip_downward=flip_downward, split=split, log_scale=log_scale)


def show_far_field(far_field, show_plots=True, show_opts=[{'label':'far_field'}], save_plots=False, save_opts=None,
                   save_data=False, data_format='hdf5', outputdir='.',
                   flip_downward=True, split=True, log_scale=False):
    """Display and export the far field.

    Args:
        far_field (smuthi.field_expansion.FarField):    Far-field object to show and export
        show_plots (bool):      Display plots if True
        show_opts (dict list):  List of dictionaries containing options to be passed to pcolormesh for plotting.
                                If save_plots=True, a 1:1 correspondence between show_opts and save_opts dictionaries
                                is assumed. For simplicity, one can also provide a single show_opts entry that will
                                be applied to all save_opts.
                                The following keys are available (see matplotlib.pyplot.pcolormesh documentation):
                                'alpha'     (None)
                                'cmap'      ('inferno')
                                'norm'      (None), is set to matplotlib.colors.LogNorm() if log_scale is True
                                'vmin'      (None), applies only to 2D plots
                                'vmax'      (None), applies only to 2D plots
                                'shading'   ('nearest'), applies only to 2D plots. 'gouraud' is also available
                                'linewidth' (None), applies only to 1D plots
                                'linestyle' (None), applies only to 1D plots
                                'marker'    (None), applies only to 1D plots
                                An optional extra key called 'label' of type string is shown in the plot title
                                and appended to the associated file if save_plots is True
                                Finally, an optional 'figsize' key is available to set the width and height of
                                the figure window (see matplotlib.pyplot.figure documentation)
        save_plots (bool):      If True, plots are exported to file.
        save_opts (dict list):  List of dictionaries containing options to be passed to savefig.
                                A 1:1 correspondence between save_opts and show_opts dictionaries is assumed. For
                                simplicity, one can also provide a single save_opts entry that will be applied to
                                all show_opts.
                                The following keys are made available (see matplotlib.pyplot.savefig documentation):
                                'dpi'           (None)
                                'orientation'   (None)
                                'format'        ('png'), also available: eps, jpeg, jpg, pdf, ps, svg, tif, tiff ...
                                'transparent'   (False)
                                'bbox_inches'   ('tight')
                                'pad_inches'    (0.1)
        save_data (bool):       If True, raw data are exported to file
        data_format (str):      Output data format string, 'hdf5' and 'ascii' formats are available
        outputdir (str):        Path to the directory where files are to be saved
        flip_downward (bool):   If True, represent downward directions as 0-90 deg instead of 90-180
        split (bool):           If True, show two different plots for upward and downward directions
        log_scale (bool):       If True, set a logarithmic scale
    """

    if split and any(far_field.polar_angles < np.pi/2) and any(far_field.polar_angles > np.pi/2):
        for d in show_opts:
            d['label'] = d.get('label') + '_top'
        show_far_field(far_field.top(), show_plots, show_opts, save_plots, save_opts,
                       save_data, data_format, outputdir, True, False, log_scale)
        for d in show_opts:
            d['label'] = d.get('label').rstrip('_top') + '_bottom'
        show_far_field(far_field.bottom(), show_plots, show_opts, save_plots, save_opts,
                       save_data, data_format, outputdir, True, False, log_scale)
        return

    if (not os.path.exists(outputdir)) and (save_plots or save_data):
        os.makedirs(outputdir)

    if save_opts is None:
        save_opts = [{'format':'png'}]

    if save_data:
        pa = far_field.polar_angles
        aa = far_field.azimuthal_angles
        label = show_opts[0].get('label') # WARNING: filenames are based only on the label in show_opts[0]
        if data_format.lower() == 'hdf5':
            import h5py

            with h5py.File(outputdir + '/data.hdf5', 'a') as f:
                g = f.require_group('far_field')
                g.require_dataset('polar_angles', data=pa, shape=np.shape(pa), dtype=pa.dtype)
                g.require_dataset('azimuthal_angles', data=aa, shape=np.shape(aa), dtype=aa.dtype)
                g.create_dataset(label, data=far_field.signal, compression="gzip")
                g.create_dataset(label + '_polar', data=far_field.azimuthal_integral(), compression="gzip")
        elif data_format.lower() == 'ascii':
            np.savetxt(outputdir + '/' + label + '_TE.dat', far_field.signal[0, :, :],
                       header='Each line corresponds to a polar angle, each column corresponds to an azimuthal angle.')
            np.savetxt(outputdir + '/' + label + '_TM.dat', far_field.signal[1, :, :],
                       header='Each line corresponds to a polar angle, each column corresponds to an azimuthal angle.')
            np.savetxt(outputdir + '/' + label + '_polar_TE.dat', far_field.azimuthal_integral()[0, :],
                       header='Each line corresponds to a polar angle, each column corresponds to an azimuthal angle.')
            np.savetxt(outputdir + '/' + label + '_polar_TM.dat', far_field.azimuthal_integral()[1, :],
                       header='Each line corresponds to a polar angle, each column corresponds to an azimuthal angle.')
            np.savetxt(outputdir + '/polar_angles.dat', pa,
                       header='Polar angles of the far field in radians.')
            np.savetxt(outputdir + '/azimuthal_angles.dat', aa,
                       header='Azimuthal angles of the far field in radians.')
        else:
            raise ValueError('Currently, only hdf5 or ascii output data formats are available')

    alpha_grid = far_field.alpha_grid()
    beta_grid = far_field.beta_grid() * 180 / np.pi
    polar_array = far_field.polar_angles * 180 / np.pi
    if flip_downward and all(far_field.polar_angles >= np.pi / 2):
        beta_grid = 180 - beta_grid
        polar_array = 180 - polar_array

    if not show_plots:
        import matplotlib
        default_backend = matplotlib.get_backend()
        matplotlib.use('Agg') # a non-gui backend to avoid opening figure windows

    for show_opt, save_opt in zip(cycle(show_opts), save_opts) if len(show_opts) < len(save_opts) else zip(show_opts, cycle(save_opts)):
        # 2D polar plot of far field
        fig = plt.figure(figsize=show_opt.get('figsize',[6.4, 4.8]))
        ax = fig.add_subplot(111, polar=True)

        if log_scale: # in either case, this will be overridden if a 'norm' is also passed to show_opts
            color_norm = show_opt.get('norm', LogNorm(vmin=show_opt.get('vmin'), vmax=show_opt.get('vmax')))
        else:
            color_norm = show_opt.get('norm', Normalize(vmin=show_opt.get('vmin'), vmax=show_opt.get('vmax')))

        pcm = ax.pcolormesh(alpha_grid, beta_grid, (far_field.signal[0, :, :] + far_field.signal[1, :, :]),
                            alpha=show_opt.get('alpha'), norm=color_norm,
                            cmap=show_opt.get('cmap','inferno'), shading=show_opt.get('shading','nearest'))

        plt.colorbar(pcm, ax=ax)
        plt.title(show_opt.get('label').replace('_',' '))
        if save_plots:
            export_filename = outputdir + '/' + show_opt.get('label') + '.' + save_opt.get('format','png') # png unless specified
            plt.savefig(export_filename, dpi=save_opt.get('dpi'), orientation=save_opt.get('orientation','portrait'),
                        transparent=save_opt.get('transparent',False), bbox_inches=save_opt.get('bbox_inches','tight'),
                        pad_inches=save_opt.get('pad_inches',0.1))

        if not show_plots:
            plt.close(fig)

        # 1D polar plot of far field
        fig = plt.figure(figsize=show_opt.get('figsize',[6.4, 4.8]))

        plt.plot(polar_array, np.sum(far_field.azimuthal_integral(), axis=0) * np.pi / 180, alpha=show_opt.get('alpha'),
                 lw=show_opt.get('linewidth'), ls=show_opt.get('linestyle'), marker=show_opt.get('marker'))

        plt.xlabel('polar angle (degree)') # TODO: use radians instead?
        if far_field.signal_type == 'differential scattering cross section':
            plt.ylabel('d_CS/d_cos(beta)') # TODO: add units based on simulation.length_unit?
        elif far_field.signal_type == 'intensity':
            plt.ylabel('d_P/d_cos(beta)') # TODO: add units based on simulation.length_unit?

        if isinstance(color_norm, LogNorm):
            plt.yscale('log')
        elif isinstance(color_norm, SymLogNorm):
            linscale = color_norm._linscale_adj*(1.0 - 1/color_norm._base)
            plt.yscale('symlog', linthresh=color_norm.linthresh,
                                 base=color_norm._base, linscale=linscale)

        # the following line would apply the same vmin and vmax of 2D maps to 1D polar plots
        # plt.ylim([color_norm.vmin, color_norm.vmax])
        plt.grid(True)
        plt.title(show_opt.get('label').replace('_',' '))

        if save_plots:
            export_filename = outputdir + '/' + show_opt.get('label') + '_polar.' + save_opt.get('format','png')
            plt.savefig(export_filename, dpi=save_opt.get('dpi'), orientation=save_opt.get('orientation','portrait'),
                        transparent=save_opt.get('transparent',False), bbox_inches=save_opt.get('bbox_inches','tight'),
                        pad_inches=save_opt.get('pad_inches',0.1))
        if not show_plots:
            plt.close(fig)

    if not show_plots:
        matplotlib.use(default_backend) # now we can restore the original backend
    else:
        plt.show(block=False)



"""Manage post processing steps to evaluate the scattered far field"""

import os
import numpy as np
import smuthi.fields as flds
import smuthi.postprocessing.scattered_field as sf


class FarField:
    r"""Represent the far field amplitude and far field intensity of an
    electromagnetic field.

    The electric field amplitude :math:`\mathbf{A}(\theta,\phi)` is defined by

    .. math::
        \mathbf{E}(\mathbf{r}) \approx \frac{e^{ikr}}{-ikr} \mathbf{A}(\theta,\phi)

    for :math:`kr\rightarrow\infty`, compare equation (3.10) of Bohren and Huffman's
    textbook on light scattering.

    In the above, :math:`\mathbf{A}(\theta,\phi)` is a complex, vector valued
    function of polar and azimuthal angle. It contains information on the amplitude
    and phase of the scattered electric field in far field domain.

    The intensity :math:`I_{\Omega,j}(\beta, \alpha)` is defined by

    .. math::
        P = \sum_{j=1}^2 \iint \mathrm{d}^2 \Omega \, I_{\Omega,j}(\beta, \alpha),

    where :math:`P` is the radiative power, :math:`j` indicates the polarization and
    :math:`\mathrm{d}^2 \Omega = \mathrm{d}\alpha \sin\beta \mathrm{d}\beta`
    denotes the infinitesimal solid angle.

    Args:
        polar_angles (numpy.ndarray):       array of polar angles for plane wave expansions. If 'default', use
                                            smuthi.fields.default_polar_angles
        azimuthal_angles (ndarray or str):  array of azimuthal angles for plane wave expansions. If 'default', use
                                            smuthi.fields.default_azimuthal_angles
        angular_resolution (float):         If provided, angular arrays are generated with this angular resolution
                                            over the default angular range
        signal_type (str):                  Use this field to describe the physical meaning of the power related signal (e.g., 'intensity' for standard power flux far fields).
        reference_point (list or tuple):    [x, y, z]-coordinates of point relative to which the far field is defined
    """

    def __init__(self, polar_angles='default', azimuthal_angles='default',
                 angular_resolution=None, signal_type='intensity', reference_point=None):
        azimuthal_angles, polar_angles = _angular_arrays(azimuthal_angles, polar_angles, angular_resolution)
        self.polar_angles = polar_angles
        self.azimuthal_angles = azimuthal_angles

        self.reference_point = reference_point

        # The far field amplitude is represented as a 3-dimensional numpy.ndarray.
        # The indices are:
        # - polarization (0=TE, 1=TM)
        # - index of the polar angle
        # - index of the azimuthal angle
        self.amplitude = np.zeros((2, len(polar_angles), len(azimuthal_angles)),
                                  dtype=complex)
        self.amplitude.fill(np.nan)

        # The far field signal is represented as a 3-dimensional numpy.ndarray.
        # The indices are:
        # - polarization (0=TE, 1=TM)
        # - index of the polar angle
        # - index of the azimuthal angle
        self.signal = np.zeros((2, len(polar_angles), len(azimuthal_angles)))
        self.signal.fill(np.nan)
        self.signal_type = signal_type

    def electric_field_amplitude(self):
        """Evaluate electric field amplitude vector

            Returns:
                Tuple of (A_x, A_y, A_z) numpy.ndarray objects with the Cartesian
                coordinates of complex electric field amplitude.
        """
        # azimuth unit vector
        e_phi_x = - np.sin(self.azimuthal_angles)[np.newaxis, :]
        e_phi_y = np.cos(self.azimuthal_angles)[np.newaxis, :]

        # polar unit vector
        e_theta_x = np.outer(np.cos(self.polar_angles), np.cos(self.azimuthal_angles))
        e_theta_y = np.outer(np.cos(self.polar_angles), np.sin(self.azimuthal_angles))
        e_theta_z = -np.sin(self.polar_angles)[:, np.newaxis]

        Ax = self.amplitude[0, :, :] * e_phi_x + self.amplitude[1, :, :] * e_theta_x
        Ay = self.amplitude[0, :, :] * e_phi_y + self.amplitude[1, :, :] * e_theta_y
        Az =                                     self.amplitude[1, :, :] * e_theta_z

        return Ax, Ay, Az

    def azimuthal_integral_times_sin_beta(self):
        r"""Far field intensity as a function of polar angle only.

        .. math::
            P = \sum_{j=1}^2 \int \mathrm{d} \beta \, I_{\beta,j}(\beta),

        with

        .. math::
            I_{\beta,j}(\beta) = \int \mathrm{d} \alpha \, \sin\beta I_j(\beta, \alpha),

        Returns:
            :math:`I_{\beta,j}(\beta)` as numpy ndarray. First index is polarization, second is polar angle.
        """
        if len(self.azimuthal_angles) > 2:
            return np.trapz(self.signal, self.azimuthal_angles[None, None, :]) * np.sin(self.polar_angles[None, :])
        else:
            return None

    def azimuthal_integral(self):
        r"""Far field intensity as a function of the polar angle cosine only.

        .. math::
            P = \sum_{j=1}^2 \int \mathrm{d} \cos\beta \, I_{\cos\beta,j}(\beta),

        with

        .. math::
            I_{\beta,j}(\beta) = \int \mathrm{d} \alpha \, I_j(\beta, \alpha),

        Returns:
            :math:`I_{\cos\beta,j}(\beta)` as numpy ndarray. First index is polarization, second is polar angle.
        """
        if len(self.azimuthal_angles) > 2:
            return np.trapz(self.signal, self.azimuthal_angles[None, None, :])
        else:
            return None

    def integral(self):
        r"""Integrate intensity to obtain total power :math:`P`.

        Returns:
            :math:`P_j` as numpy 1D-array with length 2, the index referring to polarization.
        """
        if len(self.azimuthal_angles) > 2:
            return np.trapz(self.azimuthal_integral_times_sin_beta(), self.polar_angles[None, :])
        else:
            return None

    def top(self):
        r"""Split far field into top and bottom part.

        Returns:
            FarField object with only the intensity for top hemisphere (:math:`\beta\leq\pi/2`)
        """
        if any(self.polar_angles <= np.pi / 2):
            ff = FarField(polar_angles=self.polar_angles[self.polar_angles <= np.pi / 2],
                          azimuthal_angles=self.azimuthal_angles, signal_type=self.signal_type)
            ff.signal = self.signal[:, self.polar_angles <= np.pi / 2, :]
            ff.amplitude = self.amplitude[:, self.polar_angles <= np.pi / 2, :]
            return ff
        else:
            return None

    def bottom(self):
        r"""Split far field into top and bottom part.

        Returns:
            FarField object with only the intensity for bottom hemisphere (:math:`\beta\geq\pi/2`)
        """
        if any(self.polar_angles >= np.pi / 2):
            ff = FarField(polar_angles=self.polar_angles[self.polar_angles >= np.pi / 2],
                          azimuthal_angles=self.azimuthal_angles, signal_type=self.signal_type)
            ff.signal = self.signal[:, self.polar_angles >= np.pi / 2, :]
            ff.amplitude = self.amplitude[:, self.polar_angles >= np.pi / 2, :]
            return ff
        else:
            return None

    def alpha_grid(self):
        r"""
        Returns:
            Meshgrid with :math:`\alpha` values.
        """
        agrid, _ = np.meshgrid(self.azimuthal_angles, self.polar_angles.real)
        return agrid

    def beta_grid(self):
        r"""
        Returns:
            Meshgrid with :math:`\beta` values.
        """
        _, bgrid = np.meshgrid(self.azimuthal_angles, self.polar_angles.real)
        return bgrid

    def append(self, other):
        """Combine two FarField objects with disjoint angular ranges. The other far field is appended to this one.

        Args:
            other (FarField): far field to append to this one.
        """

        if not all(self.azimuthal_angles == other.azimuthal_angles):
            raise ValueError('azimuthal angles not consistent')
        if not self.signal_type == other.signal_type:
            raise ValueError('signal type not consistent')
        if max(self.polar_angles) <= min(other.polar_angles):
            self.polar_angles = np.concatenate((self.polar_angles, other.polar_angles))
            self.signal = np.concatenate((self.signal, other.signal), 1)
            self.amplitude = np.concatenate((self.amplitude, other.amplitude), 1)
        elif min(self.polar_angles) >= max(other.polar_angles):
            self.polar_angles = np.concatenate((other.polar_angles, self.polar_angles))
            self.signal = np.concatenate((other.signal, self.signal), 1)
            self.amplitude = np.concatenate((other.amplitude, self.amplitude), 1)
        else:
            raise ValueError('far fields have overlapping polar angle domains')


def _angular_arrays(azimuthal_angles='default', polar_angles='default', angular_resolution=None):
    """ Utility function returning angular arrays based on user input """ # NOTE: if both arrays and a resolution are provided, the arrays are ignored!
    if angular_resolution is not None:
        azimuthal_angles, polar_angles = flds.angular_arrays(angular_resolution)
    else:
        if type(polar_angles) == str and polar_angles == 'default':
            polar_angles = flds.default_polar_angles
        if type(azimuthal_angles) == str and azimuthal_angles == 'default':
            azimuthal_angles = flds.default_azimuthal_angles
    return azimuthal_angles, polar_angles


def pwe_to_ff_conversion(vacuum_wavelength, plane_wave_expansion):
    """Compute the far field of a plane wave expansion object.

    Args:
        vacuum_wavelength (float):                 Vacuum wavelength in length units.
        plane_wave_expansion (PlaneWaveExpansion): Plane wave expansion to convert into far field object.

    Returns:
        A FarField object containing the far field intensity.
    """
    omega = flds.angular_frequency(vacuum_wavelength)
    k = plane_wave_expansion.k
    kp = plane_wave_expansion.k_parallel
    if plane_wave_expansion.kind == 'upgoing':
        polar_angles = np.arcsin(kp / k)
    elif plane_wave_expansion.kind == 'downgoing':
        polar_angles = np.pi - np.arcsin(kp / k)
    else:
        raise ValueError('PWE type not specified')
    if any(polar_angles.imag):
        raise ValueError('complex angles are not allowed')
    azimuthal_angles = plane_wave_expansion.azimuthal_angles

    # amplitude
    kkz = flds.k_z(k_parallel=kp, k=k) * k
    ampli = - 2 * np.pi * kkz[np.newaxis, :, np.newaxis] * plane_wave_expansion.coefficients

    # intensity
    kkz2 = flds.k_z(k_parallel=kp, k=k) ** 2 * k
    intens = (2 * np.pi ** 2 / omega * kkz2[np.newaxis, :, np.newaxis]
              * abs(plane_wave_expansion.coefficients) ** 2).real
    srt_idcs = np.argsort(polar_angles)  # reversing order in case of downgoing
    ff = FarField(polar_angles=polar_angles[srt_idcs], azimuthal_angles=azimuthal_angles,
                  reference_point=plane_wave_expansion.reference_point)
    ff.amplitude = ampli[:, srt_idcs, :]
    ff.signal = intens[:, srt_idcs, :]
    return ff


def total_far_field(initial_field, particle_list, layer_system,
                    polar_angles='default', azimuthal_angles='default',
                    angular_resolution=None, reference_point=None):
    """
    Evaluate the total far field, the initial far field and the scattered far field. Cannot be used if initial field
    is a plane wave.

    Args:
        initial_field (smuthi.initial_field.InitialField): represents the initial field
        particle_list (list):                       list of smuthi.Particle objects
        layer_system (smuthi.layers.LayerSystem):   represents the stratified medium
        polar_angles (numpy.ndarray or str):        polar angles values (radian).
                                                    if 'default', use smuthi.fields.default_polar_angles
        azimuthal_angles (numpy.ndarray or str):    azimuthal angle values (radian)
                                                    if 'default', use smuthi.fields.default_azimuthal_angles
        angular_resolution (float):                 If provided, angular arrays are generated with this angular
                                                    resolution over the default angular range
        reference_point (list or tuple):            If set to a value other than None, the far field
                                                    will be calculated with this as its reference point.
    Returns:
        A tuple of three smuthi.field_expansion.FarField objects for total, initial and scattered far field. Mind that the scattered far field
        has no physical meaning and is for illustration purposes only.
    """
    if not (type(initial_field).__name__ == 'GaussianBeam' or type(initial_field).__name__ == 'DipoleSource'
            or type(initial_field).__name__ == 'DipoleCollection'):
        raise ValueError('only for Gaussian beams and dipole sources')
    omega = initial_field.angular_frequency()
    vacuum_wavelength = initial_field.vacuum_wavelength
    azimuthal_angles, polar_angles = _angular_arrays(azimuthal_angles, polar_angles, angular_resolution)

    if any(polar_angles.imag):
        raise ValueError("complex angles not allowed in far field")

    i_top = layer_system.number_of_layers() - 1
    top_polar_angles = polar_angles[polar_angles < (np.pi / 2)]
    bottom_polar_angles = polar_angles[polar_angles > (np.pi / 2)]
    neff_top = np.sort(np.sin(top_polar_angles) * layer_system.refractive_indices[i_top])
    neff_bottom = np.sort(np.sin(bottom_polar_angles) * layer_system.refractive_indices[0])

    if len(top_polar_angles) > 1 and layer_system.refractive_indices[i_top].imag == 0:
        pwe_scat_top, _ = sf.scattered_field_pwe(vacuum_wavelength, particle_list, layer_system, i_top,
                                                 k_parallel=neff_top*omega, azimuthal_angles=azimuthal_angles,
                                                 include_direct=True, include_layer_response=True)
        pwe_in_top, _ = initial_field.plane_wave_expansion(layer_system, i_top, k_parallel_array=neff_top*omega,
                                                           azimuthal_angles_array=azimuthal_angles)
        if reference_point is not None:
            pwe_in_top.set_reference_point(reference_point)
            pwe_scat_top.set_reference_point(reference_point)

        pwe_top = pwe_scat_top + pwe_in_top
        top_far_field = pwe_to_ff_conversion(vacuum_wavelength=vacuum_wavelength, plane_wave_expansion=pwe_top)
        top_far_field_init = pwe_to_ff_conversion(vacuum_wavelength=vacuum_wavelength,
                                                  plane_wave_expansion=pwe_in_top)
        top_far_field_scat = pwe_to_ff_conversion(vacuum_wavelength=vacuum_wavelength,
                                                  plane_wave_expansion=pwe_scat_top)
    else:
        top_far_field = None
        top_far_field_init = None
        top_far_field_scat = None

    if len(bottom_polar_angles) > 1 and layer_system.refractive_indices[0].imag == 0:
        _, pwe_scat_bottom = sf.scattered_field_pwe(vacuum_wavelength, particle_list, layer_system, 0,
                                                   k_parallel=neff_bottom*omega, azimuthal_angles=azimuthal_angles,
                                                   include_direct=True, include_layer_response=True)
        _, pwe_in_bottom = initial_field.plane_wave_expansion(layer_system, 0, k_parallel_array=neff_bottom*omega,
                                                              azimuthal_angles_array=azimuthal_angles)
        if reference_point is not None:
            pwe_in_bottom.set_reference_point(reference_point)
            pwe_scat_bottom.set_reference_point(reference_point)

        pwe_bottom = pwe_scat_bottom + pwe_in_bottom
        bottom_far_field = pwe_to_ff_conversion(vacuum_wavelength=vacuum_wavelength,
                                                plane_wave_expansion=pwe_bottom)
        bottom_far_field_init = pwe_to_ff_conversion(vacuum_wavelength=vacuum_wavelength,
                                                     plane_wave_expansion=pwe_in_bottom)
        bottom_far_field_scat = pwe_to_ff_conversion(vacuum_wavelength=vacuum_wavelength,
                                                     plane_wave_expansion=pwe_scat_bottom)
    else:
        bottom_far_field = None
        bottom_far_field_init = None
        bottom_far_field_scat = None

    if top_far_field is not None:
        far_field = top_far_field
        far_field_init = top_far_field_init
        far_field_scat = top_far_field_scat
        if bottom_far_field is not None:
            far_field.append(bottom_far_field)
            far_field_init.append(bottom_far_field_init)
            far_field_scat.append(bottom_far_field_scat)
    else:
        far_field = bottom_far_field
        far_field_init = bottom_far_field_init
        far_field_scat = bottom_far_field_scat

    far_field.polar_angles = far_field.polar_angles.real
    far_field_init.polar_angles = far_field_init.polar_angles.real
    far_field_scat.polar_angles = far_field_scat.polar_angles.real
    return far_field, far_field_init, far_field_scat


def scattered_far_field(vacuum_wavelength, particle_list, layer_system,
                        polar_angles='default', azimuthal_angles='default',
                        angular_resolution=None, reference_point=None):
    """
    Evaluate the scattered far field.

    Args:
        vacuum_wavelength (float):                  in length units
        particle_list (list):                       list of smuthi.Particle objects
        layer_system (smuthi.layers.LayerSystem):   represents the stratified medium
        polar_angles (numpy.ndarray or str):        polar angles values (radian).
                                                    if 'default', use smuthi.fields.default_polar_angles
        azimuthal_angles (numpy.ndarray or str):    azimuthal angle values (radian)
                                                    if 'default', use smuthi.fields.default_azimuthal_angles
        angular_resolution (float):                 If provided, angular arrays are generated with this angular
                                                    resolution over the default angular range
        reference_point (list or tuple):            If set to a value other than None, the far field
                                                    will be calculated with this as its reference point.

    Returns:
        A smuthi.field_expansion.FarField object of the scattered field.
    """
    omega = flds.angular_frequency(vacuum_wavelength)
    azimuthal_angles, polar_angles = _angular_arrays(azimuthal_angles, polar_angles, angular_resolution)

    if any(polar_angles.imag):
        raise ValueError("complex angles not allowed in far field")

    i_top = layer_system.number_of_layers() - 1
    top_polar_angles = polar_angles[polar_angles < (np.pi / 2)]
    bottom_polar_angles = polar_angles[polar_angles > (np.pi / 2)]
    neff_top = np.sort(np.sin(top_polar_angles) * layer_system.refractive_indices[i_top])
    neff_bottom = np.sort(np.sin(bottom_polar_angles) * layer_system.refractive_indices[0])

    if len(top_polar_angles) > 1 and layer_system.refractive_indices[i_top].imag == 0:
        pwe_top, _ = sf.scattered_field_pwe(vacuum_wavelength, particle_list, layer_system, i_top,
                                            k_parallel=neff_top*omega, azimuthal_angles=azimuthal_angles,
                                            include_direct=True, include_layer_response=True)
        if reference_point is not None:
            pwe_top.set_reference_point(reference_point)
        top_far_field = pwe_to_ff_conversion(vacuum_wavelength=vacuum_wavelength, plane_wave_expansion=pwe_top)
    else:
        top_far_field = None

    if len(bottom_polar_angles) > 1 and layer_system.refractive_indices[0].imag == 0:
        _, pwe_bottom = sf.scattered_field_pwe(vacuum_wavelength, particle_list, layer_system, 0,
                                               k_parallel=neff_bottom*omega, azimuthal_angles=azimuthal_angles,
                                               include_direct=True, include_layer_response=True)
        if reference_point is not None:
            pwe_bottom.set_reference_point(reference_point)
        bottom_far_field = pwe_to_ff_conversion(vacuum_wavelength=vacuum_wavelength,
                                                plane_wave_expansion=pwe_bottom)
    else:
        bottom_far_field = None

    if top_far_field is not None:
        far_field = top_far_field
        if bottom_far_field is not None:
            far_field.append(bottom_far_field)
    else:
        far_field = bottom_far_field

    far_field.polar_angles = far_field.polar_angles.real
    return far_field


def scattering_cross_section(initial_field, particle_list, layer_system,
                             polar_angles='default', azimuthal_angles='default',
                             angular_resolution=None):
    """Evaluate and display the differential scattering cross section as a function of solid angle.

    Args:
        initial_field (smuthi.initial.PlaneWave):   Initial Plane wave
        particle_list (list):                       scattering particles
        layer_system (smuthi.layers.LayerSystem):   stratified medium
        polar_angles (numpy.ndarray or str):        polar angles values (radian).
                                                    if 'default', use smuthi.fields.default_polar_angles
        azimuthal_angles (numpy.ndarray or str):    azimuthal angle values (radian)
                                                    if 'default', use smuthi.fields.default_azimuthal_angles
        angular_resolution (float):                 If provided, angular arrays are generated with this angular
                                                    resolution over the default angular range

    Returns:
        A smuthi.field_expansion.FarField object.
    """
    if not type(initial_field).__name__ == 'PlaneWave':
        raise ValueError('Cross section only defined for plane wave excitation.')

    i_top = layer_system.number_of_layers() - 1
    vacuum_wavelength = initial_field.vacuum_wavelength
    omega = flds.angular_frequency(vacuum_wavelength)
    k_bot = omega * layer_system.refractive_indices[0]
    k_top = omega * layer_system.refractive_indices[-1]

    # read plane wave parameters
    A_P = initial_field.amplitude
    beta_P = initial_field.polar_angle
    if np.cos(beta_P) > 0:
        i_P = 0
        n_P = layer_system.refractive_indices[i_P]
    else:
        i_P = i_top
        n_P = layer_system.refractive_indices[i_P]
    if n_P.imag:
        raise ValueError('plane wave from absorbing layer: cross section undefined')
    else:
        n_P = n_P.real

    initial_intensity = abs(A_P) ** 2 * n_P / 2

    dscs = scattered_far_field(vacuum_wavelength, particle_list, layer_system,
                               polar_angles, azimuthal_angles, angular_resolution)
    dscs.signal_type = 'differential scattering cross section'
    dscs.signal = dscs.signal / initial_intensity

    return dscs

def total_scattering_cross_section(simulation=None, initial_field=None, particle_list=None, layer_system=None,
                                   polar_angles='default', azimuthal_angles='default', angular_resolution=None):
    """Evaluate the total scattering cross section.

    Args:
        simulation (smuthi.Simulation.simulation):        Simulation object (optional)
        initial_field (smuthi.initial_field.PlaneWave):   Initial Plane wave (optional)
        particle_list (list):                       scattering particles (optional)
        layer_system (smuthi.layers.LayerSystem):   stratified medium (optional)
        polar_angles (numpy.ndarray or str):        polar angles values (radian, default None).
                                                    If None, use smuthi.fields.default_polar_angles
        azimuthal_angles (numpy.ndarray or str):    azimuthal angle values (radian, default None).
                                                    If None, use smuthi.fields.default_azimuthal_angles
        angular_resolution (float):                 If provided, angular arrays are generated with this angular
                                                    resolution over the default angular range

    Returns:
        A tuple of smuthi.field_expansion.FarField objects, one for forward scattering (i.e., into the top hemisphere) and one for backward
        scattering (bottom hemisphere).
    """

    if initial_field is None:
        initial_field = simulation.initial_field

    if particle_list is None:
        particle_list = simulation.particle_list

    if layer_system is None:
        layer_system = simulation.layer_system

    azimuthal_angles, polar_angles = _angular_arrays(azimuthal_angles, polar_angles, angular_resolution)

    dscs = scattering_cross_section(initial_field, particle_list, layer_system,
                                    polar_angles, azimuthal_angles, angular_resolution)
    scs = dscs.integral()
    return scs[0] + scs[1]


def extinction_cross_section(simulation=None, initial_field=None, particle_list=None, layer_system=None,
                             only_l=None, only_m=None, only_pol=None, only_tau=None,
                             extinction_direction='both'):
    """Evaluate the extinction cross section.

    Args:
        simulation (smuthi.Simulation.simulation):  Simulation object (optional)
        initial_field (smuthi.initial_field.PlaneWave): Plane wave object (optional)
        particle_list (list): List of smuthi.particles.Particle objects (optional)
        layer_system (smuthi.layers.LayerSystem): Representing the stratified medium
        only_pol (int):  if set to 0 or 1, only this plane wave polarization (0=TE, 1=TM) is considered
        only_tau (int):  if set to 0 or 1, only this spherical vector wave polarization (0 — magnetic, 1 — electric) is
                         considered
        only_l (int):    if set to positive number, only this multipole degree is considered
        only_m (int):    if set non-negative number, only this multipole order is considered
        extinction_direction (string): if set to 'both': return full excinction,
                                       if to 'reflection': extinction of reflected wave,
                                       if to 'transmission': extinction of transmitted wave.
                                       See section on :ref:`ExtinctionCrossSectionAnchor` for details.
    Returns:
        Extinction cross section.
    """
    if initial_field is None:
        initial_field = simulation.initial_field

    if particle_list is None:
        particle_list = simulation.particle_list

    if layer_system is None:
        layer_system = simulation.layer_system

    if not type(initial_field).__name__ == 'PlaneWave':
        raise ValueError('Cross section only defined for plane wave excitation.')

    i_top = layer_system.number_of_layers() - 1
    vacuum_wavelength = initial_field.vacuum_wavelength
    omega = flds.angular_frequency(vacuum_wavelength)
    k_bot = omega * layer_system.refractive_indices[0]
    k_top = omega * layer_system.refractive_indices[-1]

    # read plane wave parameters
    pol_P = initial_field.polarization
    beta_P = initial_field.polar_angle
    alpha_P = initial_field.azimuthal_angle

    if np.cos(beta_P) > 0:
        i_P = 0
        n_P = layer_system.refractive_indices[i_P]
        k_P = k_bot
    else:
        i_P = i_top
        n_P = layer_system.refractive_indices[i_P]
        k_P = k_top
    if n_P.imag:
        raise ValueError('plane wave from absorbing layer: cross section undefined')
    else:
        n_P = n_P.real

    # complex amplitude of initial wave (including phase factor for reference point)
    kappa_P = np.sin(beta_P) * k_P
    kx = np.cos(alpha_P) * kappa_P
    ky = np.sin(alpha_P) * kappa_P
    pm_kz_P = k_P * np.cos(beta_P)
    kvec_P = np.array([kx, ky, pm_kz_P])
    rvec_iP = np.array([0, 0, layer_system.reference_z(i_P)])
    rvec_0 = np.array(initial_field.reference_point)
    ejkriP = np.exp(1j * np.dot(kvec_P, rvec_iP - rvec_0))
    A_P = initial_field.amplitude * ejkriP

    initial_intensity = abs(A_P) ** 2 * n_P / 2

    pwe_scat_top, _ = sf.scattered_field_pwe(vacuum_wavelength, particle_list, layer_system, i_top, kappa_P, alpha_P,
                                             only_l=only_l, only_m=only_m, only_pol=only_pol, only_tau=only_tau)

    _, pwe_scat_bottom = sf.scattered_field_pwe(vacuum_wavelength, particle_list, layer_system, 0, kappa_P, alpha_P,
                                                only_l=only_l, only_m=only_m, only_pol=only_pol, only_tau=only_tau)

    # bottom extinction
    _, pwe_init_bottom = initial_field.plane_wave_expansion(layer_system, 0)
    kz_bot = flds.k_z(k_parallel=kappa_P, k=k_bot)
    gRPbot =  np.squeeze(pwe_init_bottom.coefficients[pol_P])
    g_scat_bottom = np.squeeze(pwe_scat_bottom.coefficients[pol_P])
    P_bot_ext = 4 * np.pi ** 2 * kz_bot / omega * (gRPbot * np.conj(g_scat_bottom)).real
    bottom_extinction_cs = - P_bot_ext / initial_intensity

    # top extinction
    pwe_init_top, _ = initial_field.plane_wave_expansion(layer_system, i_top)
    gRPtop = np.squeeze(pwe_init_top.coefficients[pol_P])
    kz_top = flds.k_z(k_parallel=kappa_P, k=k_top)
    g_scat_top = np.squeeze(pwe_scat_top.coefficients[pol_P])
    P_top_ext = 4 * np.pi ** 2 * kz_top / omega * (gRPtop * np.conj(g_scat_top)).real
    top_extinction_cs = - P_top_ext / initial_intensity

    bottom_error = ""
    if not k_bot.imag == 0:
        bottom_error = "Extinction into bottom direction cannot be evaluated " \
                       "because bottom layer is lossy."
    elif kappa_P > k_bot:
        bottom_error = "Extinction into bottom direction cannot be evaluated " \
                       "because wave is evanescent."

    top_error = ""
    if not k_top.imag == 0:
        top_error = "Extinction into top direction cannot be evaluated " \
                       "because top layer is lossy."
    elif kappa_P > k_top:
        top_error = "Extinction into top direction cannot be evaluated " \
                    "because wave is evanescent."

    if extinction_direction == 'both':
        if not bottom_extinction_cs.imag == 0:
            raise ValueError(bottom_error)
        if not top_extinction_cs.imag == 0:
            raise ValueError(top_error)
        return bottom_extinction_cs.real + top_extinction_cs.real

    elif extinction_direction == 'reflection':
        if i_P == 0:
            if not bottom_extinction_cs.imag == 0:
                raise ValueError(bottom_error)
            return bottom_extinction_cs.real
        else:
            if not top_extinction_cs.imag == 0:
                raise ValueError(top_error)
            return top_extinction_cs.real

    elif extinction_direction == 'transmission':
        if i_P == 0:
            if not top_extinction_cs.imag == 0:
                raise ValueError(top_error)
            return top_extinction_cs.real
        else:
            if not bottom_extinction_cs.imag == 0:
                raise ValueError(bottom_error)
            return bottom_extinction_cs.real
    else:
        raise ValueError('extinction_direction can be only \'both\', \
                                        \'reflection\' or \'transmission\'')
